#!/usr/bin/env python3

"""
Conan recipe widget representation
"""

import logging
import os
import pathlib
import re
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2, PYSIDE6
import git

if PYSIDE2:
    from cruiz.pyside2.recipe_window import Ui_RecipeWindow

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.recipe_window import Ui_RecipeWindow

    QAction = QtGui.QAction

from cruiz.commands.context import ConanContext
from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.dependencygraph import dependencygraph_from_node_dependees
from cruiz.commands.logdetails import LogDetails
from cruiz.settings.managers.generalpreferences import GeneralSettingsReader
from cruiz.settings.managers.recipe import WorkflowCwd
from cruiz.settings.managers.recipe import (
    RecipeSettings,
    RecipeSettingsReader,
    RecipeSettingsWriter,
)
from cruiz.settings.managers.namedlocalcache import NamedLocalCacheSettingsReader
from cruiz.settings.managers.fontpreferences import FontSettingsReader, FontUsage
from cruiz.widgets.util import BlockSignals, clear_widgets_from_layout
from cruiz.revealonfilesystem import reveal_on_filesystem
import cruiz.workers.lockcreate
from cruiz.model.graphaslistmodel import DependenciesListModel, DependenciesTreeModel
from cruiz.manage_local_cache import ManageLocalCachesDialog
import cruiz.revealonfilesystem
import cruiz.globals
from cruiz.exceptions import RecipeInspectionError

from .recipe import Recipe
from .findtextdialog import FindTextDialog
from .expressioneditordialog import ExpressionEditorDialog
from .dependencyview import InverseDependencyViewDialog


logger = logging.getLogger(__name__)

if PYSIDE6:
    QShortcut = QtGui.QShortcut
else:
    QShortcut = QtWidgets.QShortcut


class RecipeWidget(QtWidgets.QMainWindow):
    """
    Widget representing a Conan recipe.
    """

    configuration_changed = QtCore.Signal()
    local_workflow_changed = QtCore.Signal()

    def __init__(
        self,
        path: pathlib.Path,
        uuid: QtCore.QUuid,
        cache_name: str,
        parent: typing.Optional[QtWidgets.QWidget] = None,
    ) -> None:
        super().__init__(parent)
        self._ui = Ui_RecipeWindow()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        try:
            self._git_repository: typing.Optional[git.Repo] = git.Repo(
                path, search_parent_directories=True
            )
        except git.exc.InvalidGitRepositoryError:  # type: ignore
            self._git_repository = None
        self._ui.paneSplitter.setStretchFactor(0, 4)
        self._ui.paneSplitter.setStretchFactor(1, 1)
        self._ui.outputPane.customContextMenuRequested.connect(self._pane_context_menu)
        self._ui.errorPane.customContextMenuRequested.connect(self._pane_context_menu)
        self._empty_widget = QtWidgets.QWidget(self)
        self._empty_widget.setFixedSize(0, 0)
        self._disable_delete_on_default_output_tab()
        self._ui.pane_tabs.tabCloseRequested.connect(self._on_tab_close_request)

        self._find_shortcut = QShortcut(QtGui.QKeySequence("Ctrl+F"), self)
        self._find_shortcut.setContext(QtCore.Qt.WidgetWithChildrenShortcut)
        self._find_shortcut.activated.connect(self._open_find_dialog)
        with GeneralSettingsReader() as settings:
            combine_panes = settings.combine_panes.resolve()
            use_batching = settings.use_stdout_batching.resolve()
        self.combined_output_and_error_logs = combine_panes
        if self.combined_output_and_error_logs:
            self._ui.errorPane.hide()
        self._set_pane_font()
        self._ui.actionOpen_recipe_in_editor.triggered.connect(
            self._open_recipe_in_editor
        )
        self._ui.actionOpen_recipe_folder.triggered.connect(self._open_recipe_folder)
        self._ui.actionCopy_recipe_folder_to_clipboard.triggered.connect(
            self._copy_recipe_folder_to_clipboard
        )
        self._ui.actionOpen_another_version.triggered.connect(
            self._open_another_version
        )
        self._ui.actionManage_associated_local_cache.triggered.connect(
            self._manage_associated_local_cache
        )
        self._ui.actionReload.triggered.connect(self._reload)
        self._ui.actionClose.triggered.connect(self._close_recipe)
        self._ui.localWorkflowCwd.currentIndexChanged.connect(
            self._changed_local_workflow_cwd
        )
        self._ui.localWorkflowCwd.setItemData(0, WorkflowCwd.RELATIVE_TO_RECIPE)
        self._ui.localWorkflowCwd.setItemData(1, WorkflowCwd.RELATIVE_TO_GIT)
        if not self._git_repository:
            # if not a git repository, disable the option to use relative to git
            model = self._ui.localWorkflowCwd.model()
            item = model.item(1)
            item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEnabled)  # type: ignore
        # TODO: need to fix this responsiveness
        # the QLineEdit.textChanged signal is not used here,
        # as the connected slots are so heavy weight that
        # user responsiveness on typing to the widgets is very bad
        # common subdir
        self._ui.localWorkflowCommonSubdir.editingFinished.connect(
            self._local_workflow_update_common_subdir
        )
        trash_icon = self.style().standardIcon(QtWidgets.QStyle.SP_TrashIcon)
        self._local_workflow_common_subdir_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_common_subdir_trash_action.setToolTip("Delete folder")
        self._local_workflow_common_subdir_trash_action.triggered.connect(
            self._local_workflow_on_delete_common_subdir
        )
        self._ui.localWorkflowCommonSubdir.addAction(
            self._local_workflow_common_subdir_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # install command
        self._ui.localWorkflowInstallFolder.editingFinished.connect(
            self._local_workflow_update_install_folder
        )
        self._local_workflow_install_folder_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_install_folder_trash_action.setToolTip("Delete folder")
        self._local_workflow_install_folder_trash_action.triggered.connect(
            self._local_workflow_on_delete_install_folder
        )
        self._ui.localWorkflowInstallFolder.addAction(
            self._local_workflow_install_folder_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # imports command
        self._ui.localWorkflowImportsFolder.editingFinished.connect(
            self._local_workflow_update_imports_folder
        )
        self._local_workflow_imports_folder_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_imports_folder_trash_action.setToolTip("Delete folder")
        self._local_workflow_imports_folder_trash_action.triggered.connect(
            self._local_workflow_on_delete_imports_folder
        )
        self._ui.localWorkflowImportsFolder.addAction(
            self._local_workflow_imports_folder_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # source command
        self._ui.localWorkflowSourceFolder.editingFinished.connect(
            self._local_workflow_update_source_folder
        )
        self._local_workflow_source_folder_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_source_folder_trash_action.setToolTip("Delete folder")
        self._local_workflow_source_folder_trash_action.triggered.connect(
            self._local_workflow_on_delete_source_folder
        )
        self._ui.localWorkflowSourceFolder.addAction(
            self._local_workflow_source_folder_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # build command
        self._ui.localWorkflowBuildFolder.editingFinished.connect(
            self._local_workflow_update_build_folder
        )
        self._local_workflow_build_folder_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_build_folder_trash_action.setToolTip("Delete folder")
        self._local_workflow_build_folder_trash_action.triggered.connect(
            self._local_workflow_on_delete_build_folder
        )
        self._ui.localWorkflowBuildFolder.addAction(
            self._local_workflow_build_folder_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # package command
        self._ui.localWorkflowPackageFolder.editingFinished.connect(
            self._local_workflow_update_package_folder
        )
        self._local_workflow_package_folder_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_package_folder_trash_action.setToolTip("Delete folder")
        self._local_workflow_package_folder_trash_action.triggered.connect(
            self._local_workflow_on_delete_package_folder
        )
        self._ui.localWorkflowPackageFolder.addAction(
            self._local_workflow_package_folder_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # test command
        self._ui.localWorkflowTestFolder.editingFinished.connect(
            self._local_workflow_update_test_folder
        )
        self._local_workflow_test_folder_trash_action = QAction(trash_icon, "", self)
        self._local_workflow_test_folder_trash_action.setToolTip("Delete folder")
        self._local_workflow_test_folder_trash_action.triggered.connect(
            self._local_workflow_on_delete_test_folder
        )
        self._ui.localWorkflowTestFolder.addAction(
            self._local_workflow_test_folder_trash_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        # presets
        self._ui.localWorkflowClearAll.clicked.connect(self._local_workflow_clear_all)
        self._ui.localWorkflowCommonBuildFolder.clicked.connect(
            self._local_workflow_common_build_folder
        )
        self._ui.localWorkflowProfileAndVersionBasedSubdirs.clicked.connect(
            self._local_workflow_full_spec_folders
        )
        self._ui.localWorkflowExpressionEditor.clicked.connect(
            self._local_workflow_expression_editor
        )
        rgx = QtCore.QRegularExpression(r"(^$|^@([a-zA-Z]+)\/([a-zA-Z]+)$)")
        comValidator = QtGui.QRegularExpressionValidator(rgx, self)
        self._ui.configurePkgRefNamespace.setValidator(comValidator)
        self._ui.configurePkgRefNamespace.editingFinished.connect(
            self._configure_packageref_namespace
        )
        self._ui.configurePackageId.customContextMenuRequested.connect(
            self._on_configure_packageid_context_menu
        )
        # tabify the docks on the right hand side
        self.tabifyDockWidget(
            self._ui.conanLocalWorkflowDock, self._ui.conanConfigureDock
        )
        # tabify the docks on the bottom side
        self.tabifyDockWidget(self._ui.conanLogDock, self._ui.conanCommandsDock)

        # dependency dock
        self._ui.dependentsTabs.setTabVisible(1, False)  # tree disabled
        self._ui.dependentsLog.hide()
        self._ui.dependentsLog.customContextMenuRequested.connect(
            self._dependents_log_context_menu
        )
        self._dependency_generate_log = LogDetails(
            self._ui.dependentsLog,
            None,
            True,
            False,
            None,
        )
        self._dependency_generate_context = ConanContext(
            cache_name, self._dependency_generate_log
        )
        self._dependency_generate_log.logging.connect(self._ui.dependentsLog.show)
        self._ui.dependency_rankdir.currentIndexChanged.connect(
            self._visualise_dependencies
        )

        # busy icon
        self._busy_icon = QtWidgets.QProgressBar(self)
        self._busy_icon.setTextVisible(False)
        self._busy_icon.setMinimum(0)
        self._busy_icon.setMaximum(1)

        # colours
        self._update_widget_colours_from_settings()

        # Git repository status label
        self._git_workspace_label = QtWidgets.QLabel(self)

        # signal connection
        self._ui.commandToolbar.command_started.connect(self._command_started)
        self._ui.commandToolbar.command_ended.connect(self._command_ended)
        self.configuration_changed.connect(
            self._on_configuration_update
        )  # note that this was seriously slow when using QLineEdit.textChanged signals
        self.local_workflow_changed.connect(self._on_local_workflow_update)
        self._ui.dependenciesPackageList.customContextMenuRequested.connect(
            self._dependency_list_context_menu
        )
        self._ui.behaviourToolbar.profile_changed.connect(
            self._generate_dependency_graph_from_profile_change
        )

        # associate with local caches
        self.log_details = LogDetails(
            self._ui.outputPane,
            self._ui.errorPane,
            self.combined_output_and_error_logs,
            use_batching,
            self._ui.conanLog,
        )
        self.log_details.start()
        self.recipe = Recipe(
            path,
            uuid,
            cache_name,
            self.log_details,
            self,
        )
        self.recipe.context.command_history_widget = self._ui.conanCommandHistory

        self._create_statusbar()
        self._load_local_workflow_dock()

    def post_init(self) -> None:
        """
        Post initialisation commands that need to be performed once the
        recipe is visible, or that might fail (e.g. recipe syntax error)
        """
        attributes = self.get_recipe_attributes()
        self._update_window_title(attributes)
        self._update_toolbars(attributes)
        self._load_configuration_dock(attributes)
        self._generate_dependency_graph(attributes)

    @property
    def _are_commands_running(self) -> bool:
        return self.recipe.context.is_busy or self._dependency_generate_context.is_busy

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if self._are_commands_running:
            QtWidgets.QMessageBox.warning(
                self,
                "Recipe cannot be closed",
                "Commands are still running. Cannot close recipe.",
            )
            event.ignore()
            return
        self._dependency_generate_log.stop()
        self._dependency_generate_context.close()
        self.log_details.stop()
        self.recipe.close()
        super().closeEvent(event)

    def get_recipe_attributes(
        self,
    ) -> typing.Dict[str, typing.Any]:
        """
        Get the attributes written into the recipe
        """
        return self.recipe.context.inspect_recipe(
            self.recipe.path, propagate_errors=True
        )

    def _get_options_from_recipe(
        self, attrs: typing.Dict[str, str]
    ) -> typing.List[typing.Tuple[str, typing.List[typing.Any], typing.Any]]:
        options = attrs.get("options", None)
        if not options:
            return []
        default_options = attrs["default_options"]
        values: typing.List[typing.Tuple[str, typing.List[typing.Any], typing.Any]] = []
        assert isinstance(options, dict)
        for key, value in options.items():
            values.append(
                (key, value, default_options[key] if default_options else None)
            )
        return values

    # TODO: why does this have to be PurePosixPath? to create such a path,
    # it cannot be pure
    def get_working_dir(
        self,
        workflow_cwd: WorkflowCwd,
        common_subdir: typing.Optional[str],
    ) -> pathlib.PurePosixPath:
        """
        Get the working directory for the recipe
        """
        if common_subdir is not None:
            common_subdir_path = pathlib.Path(common_subdir)
            if common_subdir_path.is_absolute():
                return pathlib.PurePosixPath(common_subdir_path)
        if workflow_cwd == WorkflowCwd.RELATIVE_TO_RECIPE:
            cwd = pathlib.PurePosixPath(self.recipe.folder)
        elif workflow_cwd == WorkflowCwd.RELATIVE_TO_GIT:
            assert self._git_repository
            assert self._git_repository.working_tree_dir
            cwd = pathlib.PurePosixPath(self._git_repository.working_tree_dir)
        if common_subdir is not None:
            cwd /= common_subdir
        return cwd

    def cwd_is_relative_to_recipe(self, workflow_cmd: WorkflowCwd) -> bool:
        """
        Is the current working directory relative to the recipe folder?
        """
        if workflow_cmd == WorkflowCwd.RELATIVE_TO_RECIPE:
            return True
        return False

    def tokens(self) -> typing.Tuple[typing.Dict[str, str], str]:
        """
        Get a dict of tokens usable in expressions
        """
        recipe_attributes = self.get_recipe_attributes()
        with RecipeSettingsReader.from_recipe(self.recipe) as settings_recipe:
            profile = settings_recipe.profile.resolve()
        profile_meta = self.recipe.context.get_profile_meta(profile)
        if os.path.isabs(profile):
            with NamedLocalCacheSettingsReader(
                self.recipe.context.cache_name
            ) as settings_localcache:
                extra_profile_dirs = (
                    settings_localcache.extra_profile_directories.resolve()
                )
            extra_profile_dir_name = [
                x for x in extra_profile_dirs if profile.startswith(x)
            ]
            assert len(extra_profile_dir_name) == 1
            profile_prefix = f"{extra_profile_dir_name[0]}-"
        else:
            profile_prefix = ""

        tokens = {
            "name": recipe_attributes["name"],
            # version is either in the attributes, or for version agnostic recipes,
            # check the recipe object
            "version": recipe_attributes["version"] or self.recipe.version,
            "profile": f"{profile_prefix}{os.path.basename(profile)}",
            "build_type": profile_meta["settings"]["build_type"],
            "build_type_lc": profile_meta["settings"]["build_type"].lower(),
        }
        return tokens, r"(\${(.*?)})"

    def resolve_expression(
        self, expression: typing.Optional[str]
    ) -> typing.Optional[str]:
        """
        Resolve an expression using token expansion.
        """
        if not expression:
            return None
        tokens, token_regex = self.tokens()
        matches = re.findall(token_regex, expression)
        for match, token in matches:
            if token in tokens:
                expression = expression.replace(match, tokens[token])
            else:
                raise RuntimeError(
                    f"Unrecognised macro '{token}' found in '{expression}'"
                )
        return expression

    def _update_window_title(self, attributes: typing.Dict[str, str]) -> None:
        version_in_recipe = attributes.get("version", None)
        if version_in_recipe is not None:
            self._ui.actionOpen_another_version.setEnabled(False)
        version = version_in_recipe or self.recipe.version
        name = attributes.get("name", None)
        if self.recipe.user:
            package = f"{name}/{version}@{self.recipe.user}/{self.recipe.channel}"
        else:
            package = f"{name}/{version}"
        self.setWindowTitle(
            f"Package: {package} - Local cache: {self.recipe.context.cache_name}"
        )

    def _update_toolbars(
        self, attributes: typing.Dict[str, typing.Optional[str]]
    ) -> None:
        self._ui.behaviourToolbar.refresh_content()
        self._ui.buildFeaturesToolbar.refresh_content(self.recipe.uuid)
        try:
            self._ui.commandToolbar.configure_actions()  # this is a one off
            self._ui.commandToolbar.refresh_action_shortcuts_and_tooltips(attributes)
        except Exception as exc:
            logging.debug("Updating toolbars suppressed: '%s'", str(exc))
            self._ui.commandToolbar.disable_all_actions()

    def _create_statusbar(self) -> None:
        self._ui.statusbar.addWidget(self._busy_icon)
        if self._git_repository:
            self._git_workspace_label.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self._git_workspace_label.customContextMenuRequested.connect(
                self._on_git_context_menu
            )
            self._ui.statusbar.addPermanentWidget(self._git_workspace_label)
        self._refresh_statusbar()

    def _refresh_statusbar(self) -> None:
        if self._git_repository:
            head_sha = self._git_repository.head.object.hexsha
            try:
                branch = self._git_repository.active_branch
                tracking_branch = branch.tracking_branch()
                if tracking_branch is None:
                    tooltip = "Untracked branch"
            except TypeError:
                branch = head_sha
                tracking_branch = None
                tooltip = "Detached head"
            message = f"{self._git_repository.working_tree_dir} " f"({branch})"
            if tracking_branch:
                commits_behind = self._git_repository.iter_commits(
                    f"{branch}..{tracking_branch}"
                )
                num_commits_behind = sum(1 for _ in commits_behind)
                tooltip = f"Tracking {tracking_branch}"
                if num_commits_behind > 0:
                    message += f" \u2193{num_commits_behind}"
                    tooltip += f" {num_commits_behind} commits behind"
                commits_ahead = self._git_repository.iter_commits(
                    f"{tracking_branch}..{branch}"
                )
                num_commits_ahead = sum(1 for _ in commits_ahead)
                if num_commits_ahead > 0:
                    message += f" \u2191{num_commits_ahead}"
                    tooltip += f" {num_commits_ahead} commits ahead"
                tooltip += f"\ncommit {head_sha}"

            self._git_workspace_label.setText(message)
            if tooltip:
                self._git_workspace_label.setToolTip(tooltip)

    def _on_git_context_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        fetch_action = QAction("Fetch", self)
        fetch_action.triggered.connect(self._fetch_git_repository)
        menu.addAction(fetch_action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _fetch_git_repository(self) -> None:
        assert self._git_repository
        for remote in self._git_repository.remotes:
            remote.fetch()
        self._refresh_statusbar()

    def on_preferences_update(self) -> None:
        """
        Slot called when preferences are updated
        """
        attributes = self.get_recipe_attributes()
        self._ui.commandToolbar.refresh_action_shortcuts_and_tooltips(attributes)
        self._ui.buildFeaturesToolbar.refresh_content(self.recipe.uuid)
        self._set_pane_font()
        self._update_widget_colours_from_settings()

    def _on_local_workflow_update(self) -> None:
        attributes = self.get_recipe_attributes()
        self._ui.commandToolbar.refresh_action_shortcuts_and_tooltips(attributes)

    def _on_configuration_update(self) -> None:
        self._on_local_workflow_update()
        attributes = self.get_recipe_attributes()
        self._generate_dependency_graph(attributes)
        self._update_window_title(attributes)

    def _changed_local_workflow_cwd(self, index: int) -> None:
        settings = RecipeSettings()
        settings.local_workflow_cwd = self.sender().itemData(index)
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()

    def _load_local_workflow_dock(self) -> None:
        with RecipeSettingsReader.from_recipe(self.recipe) as settings:
            workflow_cwd = settings.local_workflow_cwd.resolve()
            common_subdir = settings.local_workflow_common_subdir.resolve()
            install_folder = settings.local_workflow_install_folder.resolve()
            imports_folder = settings.local_workflow_imports_folder.resolve()
            source_folder = settings.local_workflow_source_folder.resolve()
            build_folder = settings.local_workflow_build_folder.resolve()
            package_folder = settings.local_workflow_package_folder.resolve()
            test_folder = settings.local_workflow_test_folder.resolve()

        # populate local workflow dock
        with BlockSignals(self._ui.localWorkflowCwd) as blocked_widget:
            blocked_widget.setCurrentIndex(workflow_cwd)
        with BlockSignals(self._ui.localWorkflowCommonSubdir) as blocked_widget:
            blocked_widget.setText(common_subdir)
            self._local_workflow_common_subdir_trash_action.setEnabled(
                bool(common_subdir)
            )
        with BlockSignals(self._ui.localWorkflowInstallFolder) as blocked_widget:
            blocked_widget.setText(install_folder)
            self._local_workflow_install_folder_trash_action.setEnabled(
                bool(install_folder)
            )
        with BlockSignals(self._ui.localWorkflowImportsFolder) as blocked_widget:
            blocked_widget.setText(imports_folder)
            self._local_workflow_imports_folder_trash_action.setEnabled(
                bool(imports_folder)
            )
        with BlockSignals(self._ui.localWorkflowSourceFolder) as blocked_widget:
            blocked_widget.setText(source_folder)
            self._local_workflow_source_folder_trash_action.setEnabled(
                bool(source_folder)
            )
        with BlockSignals(self._ui.localWorkflowBuildFolder) as blocked_widget:
            blocked_widget.setText(build_folder)
            self._local_workflow_build_folder_trash_action.setEnabled(
                bool(build_folder)
            )
        with BlockSignals(self._ui.localWorkflowPackageFolder) as blocked_widget:
            blocked_widget.setText(package_folder)
            self._local_workflow_package_folder_trash_action.setEnabled(
                bool(package_folder)
            )
        with BlockSignals(self._ui.localWorkflowTestFolder) as blocked_widget:
            blocked_widget.setText(test_folder)
            self._local_workflow_test_folder_trash_action.setEnabled(bool(test_folder))

    def _load_configuration_dock(
        self, recipe_attributes: typing.Dict[str, str]
    ) -> None:
        with RecipeSettingsReader.from_recipe(self.recipe) as settings:
            options = settings.options.resolve()
            attributes = settings.attribute_overrides.resolve()
        if "user" in attributes:
            assert "channel" in attributes
            with BlockSignals(self._ui.configurePkgRefNamespace) as blocked_widget:
                blocked_widget.setText(f"@{attributes['user']}/{attributes['channel']}")
        clear_widgets_from_layout(self._ui.optionsLayout)
        recipe_options = self._get_options_from_recipe(recipe_attributes)
        for i, (key, values, default_value) in enumerate(recipe_options):
            self._ui.optionsLayout.addWidget(QtWidgets.QLabel(key), i, 0)
            if isinstance(values, str) and values == "ANY":
                value_edit = QtWidgets.QLineEdit()
                if key in options:
                    value_edit.setText(options[key])
                else:
                    value_edit.setText(default_value or "")
                value_edit.editingFinished.connect(self._option_any_changed)
                self._ui.optionsLayout.addWidget(value_edit, i, 1)
            else:
                assert isinstance(values, list)
                value_combo = QtWidgets.QComboBox()
                for j, value in enumerate(values):
                    if value == default_value:
                        value_combo.addItem(f"{value} (default)")
                    else:
                        value_combo.addItem(str(value))
                    value_combo.setItemData(j, value)
                if key in options:
                    # from options from QSettings, these are all strings and
                    # will never be the default value
                    current_index = value_combo.findText(options[key])
                else:
                    if default_value is not None:
                        current_index = values.index(default_value)
                    else:
                        current_index = -1
                value_combo.setCurrentIndex(current_index)
                value_combo.currentIndexChanged.connect(self._option_combo_changed)
                self._ui.optionsLayout.addWidget(value_combo, i, 1)
        self._ui.configureOptionsBox.setVisible(bool(recipe_options))

    def _option_any_changed(self) -> None:
        # divide by 2 as they are added in pairs
        widget = self.sender()
        assert isinstance(widget, QtWidgets.QWidget)
        index_of_option = self._ui.optionsLayout.indexOf(widget) // 2
        # get the label text (the option name)
        name_of_option = (
            self._ui.optionsLayout.itemAtPosition(index_of_option, 0).widget().text()
        )
        settings = RecipeSettings()
        text = widget.text()
        settings.append_options({name_of_option: text or None})  # type: ignore
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.configuration_changed.emit()

    def _option_combo_changed(self, index: int) -> None:
        # divide by 2 as they are added in pairs
        widget = self.sender()
        assert isinstance(widget, QtWidgets.QWidget)
        index_of_option = self._ui.optionsLayout.indexOf(widget) // 2
        # get the label text (the option name)
        name_of_option = (
            self._ui.optionsLayout.itemAtPosition(index_of_option, 0).widget().text()
        )
        option_value = widget.itemText(index)
        real_option_value = widget.itemData(index)
        settings = RecipeSettings()
        if " (default)" in option_value:
            settings.append_options({name_of_option: None})  # type: ignore
        else:
            settings.append_options(
                {name_of_option: str(real_option_value)}  # type: ignore
            )
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.configuration_changed.emit()

    def _local_workflow_update_common_subdir(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_common_subdir = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_common_subdir_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_update_install_folder(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_install_folder = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_install_folder_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_update_imports_folder(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_imports_folder = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_imports_folder_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_update_source_folder(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_source_folder = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_source_folder_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_update_build_folder(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_build_folder = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_build_folder_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_update_package_folder(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_package_folder = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_package_folder_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_update_test_folder(self) -> None:
        settings = RecipeSettings()
        settings.local_workflow_test_folder = self.sender().text()
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()
        self._local_workflow_test_folder_trash_action.setEnabled(
            bool(self.sender().text())
        )

    def _local_workflow_sync_dirs(self) -> None:
        # note that this is only necessary since we couldn't use the
        # QLineEdit.textChanged signal due to significant overhead
        settings = RecipeSettings()
        settings.local_workflow_common_subdir = (
            self._ui.localWorkflowCommonSubdir.text()  # type: ignore[assignment]
        )
        self._local_workflow_common_subdir_trash_action.setEnabled(
            bool(self._ui.localWorkflowCommonSubdir)
        )
        settings.local_workflow_install_folder = (
            self._ui.localWorkflowInstallFolder.text()  # type: ignore[assignment]
        )
        self._local_workflow_install_folder_trash_action.setEnabled(
            bool(self._ui.localWorkflowInstallFolder)
        )
        settings.local_workflow_imports_folder = (
            self._ui.localWorkflowImportsFolder.text()  # type: ignore[assignment]
        )
        self._local_workflow_imports_folder_trash_action.setEnabled(
            bool(self._ui.localWorkflowImportsFolder)
        )
        settings.local_workflow_source_folder = (
            self._ui.localWorkflowSourceFolder.text()  # type: ignore[assignment]
        )
        self._local_workflow_source_folder_trash_action.setEnabled(
            bool(self._ui.localWorkflowSourceFolder)
        )
        settings.local_workflow_build_folder = (
            self._ui.localWorkflowBuildFolder.text()  # type: ignore[assignment]
        )
        self._local_workflow_build_folder_trash_action.setEnabled(
            bool(self._ui.localWorkflowBuildFolder)
        )
        settings.local_workflow_package_folder = (
            self._ui.localWorkflowPackageFolder.text()  # type: ignore[assignment]
        )
        self._local_workflow_package_folder_trash_action.setEnabled(
            bool(self._ui.localWorkflowPackageFolder)
        )
        settings.local_workflow_test_folder = (
            self._ui.localWorkflowTestFolder.text()  # type: ignore[assignment]
        )
        self._local_workflow_test_folder_trash_action.setEnabled(
            bool(self._ui.localWorkflowTestFolder)
        )
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.local_workflow_changed.emit()

    def _local_workflow_on_delete_common_subdir(self) -> None:
        with RecipeSettingsReader.from_recipe(self.recipe) as settings:
            workflow_cwd = settings.local_workflow_cwd.resolve()
            common_subdir = settings.local_workflow_common_subdir.resolve()
        assert common_subdir
        working_dir = self.get_working_dir(workflow_cwd, common_subdir)
        # working_dir may be an expression so...
        working_dir = self.resolve_expression(str(working_dir))
        result = QtWidgets.QMessageBox.question(
            self,
            "Delete directory",
            f"Delete the directory '{working_dir}'?",
        )
        if result == QtWidgets.QMessageBox.StandardButton.No:
            return
        QtCore.QDir(working_dir).removeRecursively()

    def _local_workflow_on_delete_command_folder(self, property_name: str) -> None:
        with RecipeSettingsReader.from_recipe(self.recipe) as settings:
            workflow_cwd = settings.local_workflow_cwd.resolve()
            common_subdir = settings.local_workflow_common_subdir.resolve()
            command_subdir = getattr(settings, property_name).resolve()
        assert command_subdir
        command_subdir = pathlib.Path(command_subdir)
        if command_subdir.is_absolute():
            command_dir = command_subdir
        else:
            working_dir = self.get_working_dir(workflow_cwd, common_subdir)
            command_dir = working_dir / command_subdir
        # command_dir may be an expression so...
        command_dir = self.resolve_expression(str(command_dir))
        result = QtWidgets.QMessageBox.question(
            self,
            "Delete directory",
            f"Delete the directory '{command_dir}'?",
        )
        if result == QtWidgets.QMessageBox.StandardButton.No:
            return
        QtCore.QDir(command_dir).removeRecursively()

    def _local_workflow_on_delete_install_folder(self) -> None:
        self._local_workflow_on_delete_command_folder("local_workflow_install_folder")

    def _local_workflow_on_delete_imports_folder(self) -> None:
        self._local_workflow_on_delete_command_folder("local_workflow_imports_folder")

    def _local_workflow_on_delete_source_folder(self) -> None:
        self._local_workflow_on_delete_command_folder("local_workflow_source_folder")

    def _local_workflow_on_delete_build_folder(self) -> None:
        self._local_workflow_on_delete_command_folder("local_workflow_build_folder")

    def _local_workflow_on_delete_package_folder(self) -> None:
        self._local_workflow_on_delete_command_folder("local_workflow_package_folder")

    def _local_workflow_on_delete_test_folder(self) -> None:
        self._local_workflow_on_delete_command_folder("local_workflow_test_folder")

    def _local_workflow_clear_all(self) -> None:
        self._ui.localWorkflowCommonSubdir.clear()
        self._ui.localWorkflowInstallFolder.clear()
        self._ui.localWorkflowImportsFolder.clear()
        self._ui.localWorkflowSourceFolder.clear()
        self._ui.localWorkflowBuildFolder.clear()
        self._ui.localWorkflowPackageFolder.clear()
        self._ui.localWorkflowTestFolder.clear()
        self._local_workflow_sync_dirs()

    def _local_workflow_common_build_folder(self) -> None:
        self._ui.localWorkflowCommonSubdir.setText("build")
        self._ui.localWorkflowInstallFolder.clear()
        self._ui.localWorkflowImportsFolder.clear()
        self._ui.localWorkflowSourceFolder.clear()
        self._ui.localWorkflowBuildFolder.clear()
        self._ui.localWorkflowPackageFolder.clear()
        self._ui.localWorkflowTestFolder.clear()
        self._local_workflow_sync_dirs()

    def _local_workflow_full_spec_folders(self) -> None:
        self._ui.localWorkflowCommonSubdir.setText("build")
        self._ui.localWorkflowInstallFolder.setText(
            "${name}/${version}/${profile}/install"
        )
        self._ui.localWorkflowImportsFolder.setText(
            "${name}/${version}/${profile}/imports"
        )
        self._ui.localWorkflowSourceFolder.setText("${name}/${version}/src")
        self._ui.localWorkflowBuildFolder.setText("${name}/${version}/${profile}/build")
        self._ui.localWorkflowPackageFolder.setText(
            "${name}/${version}/${profile}/package"
        )
        self._ui.localWorkflowTestFolder.setText("${name}/${version}/${profile}/test")
        self._local_workflow_sync_dirs()

    def _local_workflow_expression_editor(self) -> None:
        ExpressionEditorDialog(self).exec_()

    def _configure_packageref_namespace(self) -> None:
        # Note: this does appear to get called twice, which also appears to be a bug
        # https://stackoverflow.com/questions/26782211/qlineedit-editingfinished-signal-twice-when-changing-focus
        text = self.sender().text()
        settings = RecipeSettings()
        if text:
            regex = self.sender().validator().regularExpression()
            matches = regex.match(text)
            assert 3 == matches.lastCapturedIndex()
            user = matches.captured(2)
            channel = matches.captured(3)
            settings.append_attribute(
                {"user": user, "channel": channel}  # type: ignore
            )
        else:
            settings.append_attribute({"user": None, "channel": None})  # type: ignore
        RecipeSettingsWriter.from_recipe(self.recipe).sync(settings)
        self.configuration_changed.emit()

    def _open_recipe_in_editor(self) -> None:
        # TODO: make the editor a preference
        with GeneralSettingsReader() as settings:
            recipe_editor = settings.default_recipe_editor.resolve()
        if recipe_editor:
            script_args = [str(self.recipe.path)]
            QtCore.QProcess.execute(recipe_editor, script_args)
        else:
            url = QtCore.QUrl.fromLocalFile(str(self.recipe.path))
            QtGui.QDesktopServices.openUrl(url)

    def _open_recipe_folder(self) -> None:
        reveal_on_filesystem(str(self.recipe.folder))

    def _copy_recipe_folder_to_clipboard(self) -> None:
        QtWidgets.QApplication.clipboard().setText(str(self.recipe.folder))

    def _open_another_version(self) -> None:
        # TODO: note the last argument to this is the wrong type
        cruiz.globals.get_main_window().load_recipe(self.recipe.path, None, self)

    def _manage_associated_local_cache(self) -> None:
        dialog = ManageLocalCachesDialog(self, self.recipe.context.cache_name)
        dialog.cache_changed.connect(self.on_local_cache_changed)
        dialog.exec_()

    def _generate_dependency_graph_from_profile_change(self, profile: str) -> None:
        attributes = self.get_recipe_attributes()
        self._generate_dependency_graph(attributes)

    def _generate_dependency_graph(
        self, recipe_attributes: typing.Dict[str, str]
    ) -> None:
        # reset views
        self._ui.configurePackageId.setText("Calculating...")
        # https://stackoverflow.com/questions/46630185/qt-remove-model-from-view
        self._ui.dependenciesPackageList.setModel(None)  # type: ignore[arg-type]
        self._ui.dependencyView.clear()
        # calculate lock file
        params = CommandParameters("lock create", cruiz.workers.lockcreate.invoke)
        params.recipe_path = self.recipe.path
        params.name = recipe_attributes.get("name", None)
        params.version = self.recipe.version
        params.user = self.recipe.user
        params.channel = self.recipe.channel
        with RecipeSettingsReader.from_recipe(self.recipe) as settings:
            params.profile = settings.profile.resolve()
            for key, value in settings.options.resolve().items():
                # TODO: is this the most efficient algorithm?
                params.add_option(params.name, key, value)  # type: ignore
        self._dependency_generate_context.conancommand(
            params,
            self._dependency_generate_log,
            None,
            self._on_dependency_graph_generated,
        )

    def _on_dependency_graph_generated(
        self, payload: typing.Any, exception: typing.Any
    ) -> None:
        if payload:
            self._dependency_graph = payload
            self._ui.configurePackageId.setText(self._dependency_graph.root.package_id)
            try:
                self._visualise_dependencies(self._ui.dependency_rankdir.currentIndex())
            except FileNotFoundError as exc:
                exception = exc
                # fall through
        if exception:
            if payload is None:
                self._ui.configurePackageId.setText("Failed")
            lines = str(exception).splitlines()
            html = ""
            for line in lines:
                stripped_line = line.lstrip()
                num_leading_spaces = len(line) - len(stripped_line)
                html += "&nbsp;" * num_leading_spaces + stripped_line + "<br>"
            self._dependency_generate_log.stderr(
                f"Exception raised from running command:<br>"
                f"<font color='red'>{html}</font><br>"
            )

    def _visualise_dependencies(self, rank_dir_index: int) -> None:
        # list visualisation of dependencies
        self._dependencies_list_model = DependenciesListModel(self._dependency_graph)
        self._ui.dependenciesPackageList.setModel(self._dependencies_list_model)
        # tree visualisation of dependencies (DISABLED)
        self._dependencies_tree_model = DependenciesTreeModel(self._dependency_graph)
        self._ui.dependenciesPackageTree.setModel(self._dependencies_tree_model)
        # graphical visualisation of dependencies
        self._ui.dependencyView.visualise(self._dependency_graph, rank_dir_index)

    def _set_pane_font(self) -> None:
        with FontSettingsReader(FontUsage.OUTPUT) as settings:
            name = settings.name.resolve()
            size = settings.size.resolve()
        if name:
            font = QtGui.QFont(name, size)
        else:
            font = QtGui.QFontDatabase.systemFont(QtGui.QFontDatabase.FixedFont)
        for i in range(self._ui.pane_tabs.count()):
            tab_widget = self._ui.pane_tabs.widget(i)
            splitter = (
                tab_widget
                if isinstance(tab_widget, QtWidgets.QSplitter)
                else tab_widget.findChild(QtWidgets.QSplitter)
            )
            splitter.widget(0).document().setDefaultFont(font)
            splitter.widget(1).document().setDefaultFont(font)

    def failed_to_load(self) -> None:
        """
        Call this in a recipe failure to load situation, that disables everything
        """
        self._ui.conanDependencyDock.setEnabled(False)
        self._ui.conanCommandsDock.setEnabled(False)
        self._ui.conanCommandHistory.setEnabled(False)
        self._ui.conanConfigureDock.setEnabled(False)
        self._ui.conanLocalWorkflowDock.setEnabled(False)
        self._ui.behaviourToolbar.setEnabled(False)
        self._ui.buildFeaturesToolbar.setEnabled(False)
        self._ui.commandToolbar.disable_all_actions()
        self.setWindowTitle(f"Failed to load recipe {self.recipe.path}")

    def _command_started(self) -> None:
        self._busy_icon.setMaximum(0)
        self._ui.conanConfigureDock.setEnabled(False)
        self._ui.conanLocalWorkflowDock.setEnabled(False)
        self._ui.behaviourToolbar.setEnabled(False)
        self._ui.buildFeaturesToolbar.setEnabled(False)

    def _command_ended(self) -> None:
        self._busy_icon.setMaximum(1)
        self._ui.conanConfigureDock.setEnabled(True)
        self._ui.conanLocalWorkflowDock.setEnabled(True)
        self._ui.behaviourToolbar.setEnabled(True)
        self._ui.buildFeaturesToolbar.setEnabled(True)

    def _pane_context_menu(self, position: QtCore.QPoint) -> None:
        menu = self.sender().createStandardContextMenu(position)
        menu.addSeparator()
        find_action = QAction("Find...", self)
        find_action.setShortcut(self._find_shortcut.key())
        find_action.setShortcutVisibleInContextMenu(True)
        find_action.setData(self.sender())
        find_action.triggered.connect(self._open_find_dialog)
        menu.addAction(find_action)
        menu.addSeparator()
        clear_action = QAction("Clear", self)
        clear_action.triggered.connect(self.sender().clear)
        menu.addAction(clear_action)
        menu.addSeparator()
        pin_action = QAction("Pin to tab", self)
        pin_action.triggered.connect(self._pin_current_output)
        pin_action.setEnabled(self._ui.pane_tabs.count() == 1)
        menu.addAction(pin_action)
        menu.exec_(self.sender().viewport().mapToGlobal(position))

    def _open_find_dialog(self) -> None:
        if isinstance(self.sender(), QShortcut):
            pane = QtWidgets.QApplication.focusWidget()
        else:
            pane = self.sender().data()
        if isinstance(pane, QtWidgets.QPlainTextEdit):
            dialog = FindTextDialog(pane)
            dialog.search_forwards.connect(self._find_next)
            dialog.search_backwards.connect(self._find_prev)
            dialog.exec_()

    def _find_next(
        self,
        pane: QtWidgets.QPlainTextEdit,
        pattern: str,
        case_sensitive: bool,
        wrap_around: bool,
    ) -> None:
        flags = QtGui.QTextDocument.FindFlags()
        if case_sensitive:
            flags |= QtGui.QTextDocument.FindCaseSensitively  # type: ignore
        result = pane.find(pattern, flags)
        if not result and wrap_around:
            pane.moveCursor(QtGui.QTextCursor.Start)
            result = pane.find(pattern, flags)

    def _find_prev(
        self,
        pane: QtWidgets.QPlainTextEdit,
        pattern: str,
        case_sensitive: bool,
        wrap_around: bool,
    ) -> None:
        flags = QtGui.QTextDocument.FindBackward
        if case_sensitive:
            flags |= QtGui.QTextDocument.FindCaseSensitively  # type: ignore
        result = pane.find(pattern, flags)
        if not result and wrap_around:
            pane.moveCursor(QtGui.QTextCursor.End)
            result = pane.find(pattern, flags)

    def _dependency_list_context_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        open_package_dir_action = QAction("Open package directory", self)
        open_package_dir_action.triggered.connect(self._open_package_directory)
        menu.addAction(open_package_dir_action)
        copy_package_dir_action = QAction("Copy package directory to clipboard", self)
        copy_package_dir_action.triggered.connect(self._copy_package_directory)
        menu.addAction(copy_package_dir_action)
        what_uses_this_action = QAction("What uses this?...", self)
        what_uses_this_action.triggered.connect(self._on_what_uses_this)
        menu.addAction(what_uses_this_action)
        menu.exec_(self._ui.dependenciesPackageList.mapToGlobal(position))

    def _get_package_directory_of_current_dependency(self) -> str:
        index = self._ui.dependenciesPackageList.currentIndex()
        node = index.data(QtCore.Qt.UserRole)
        return self.recipe.context.get_package_directory(node)

    def _open_package_directory(self) -> None:
        directory = self._get_package_directory_of_current_dependency()
        if os.path.isdir(directory):
            cruiz.revealonfilesystem.reveal_on_filesystem(directory)
        else:
            QtWidgets.QMessageBox.critical(
                self,
                "No such package directory",
                f"Package directory '{directory}' does not exist",
            )

    def _copy_package_directory(self) -> None:
        directory = self._get_package_directory_of_current_dependency()
        QtWidgets.QApplication.clipboard().setText(directory)

    def _on_what_uses_this(self) -> None:
        index = self._ui.dependenciesPackageList.currentIndex()
        node = index.data(QtCore.Qt.UserRole)
        new_graph = dependencygraph_from_node_dependees(node)
        InverseDependencyViewDialog(new_graph).exec_()

    def _dependents_log_context_menu(self, position: QtCore.QPoint) -> None:
        menu = self.sender().createStandardContextMenu(position)
        menu.addSeparator()
        clear_action = QAction("Clear", self)
        clear_action.triggered.connect(self._clear_dependents_log)
        menu.addAction(clear_action)
        menu.exec_(self.sender().viewport().mapToGlobal(position))

    def _clear_dependents_log(self) -> None:
        self._ui.dependentsLog.clear()
        self._ui.dependentsLog.hide()

    def _update_widget_colours_from_settings(self) -> None:
        def _set_widget_highlight_colour(
            widget: QtWidgets.QWidget, colour: QtGui.QColor
        ) -> None:
            palette = widget.palette()
            palette.setColor(QtGui.QPalette.Highlight, colour)
            widget.setPalette(palette)

        with GeneralSettingsReader() as settings:
            busy_icon_color = settings.busy_icon_colour.resolve()
            found_text_background_color = (
                settings.found_text_background_colour.resolve()
            )

        colour = (
            f"{busy_icon_color.red()}, "
            f"{busy_icon_color.green()}, "
            f"{busy_icon_color.blue()}"
        )
        style = (
            "QProgressBar::chunk "
            f"{{background-color: rgb({colour}); width: 5px; margin: 2.5px;}}"
        )
        self._busy_icon.setStyleSheet(style)

        _set_widget_highlight_colour(self._ui.outputPane, found_text_background_color)
        _set_widget_highlight_colour(self._ui.errorPane, found_text_background_color)

    def _pin_current_output(self) -> None:
        # create a pinned readonly copy of the current output for inspection
        splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical, self)
        output = QtWidgets.QPlainTextEdit()
        output.appendHtml(self._ui.outputPane.document().toHtml())
        error = QtWidgets.QPlainTextEdit()
        error.appendHtml(self._ui.errorPane.document().toHtml())
        splitter.addWidget(output)
        splitter.addWidget(error)
        with GeneralSettingsReader() as settings:
            if settings.combine_panes.resolve():
                error.hide()
        self._ui.pane_tabs.addTab(splitter, "Pinned output")
        self._set_pane_font()

    def _disable_delete_on_default_output_tab(self) -> None:
        # default tab containing the output pane is not closable (but others are)
        self._ui.pane_tabs.tabBar().setTabButton(
            0, QtWidgets.QTabBar.LeftSide, self._empty_widget
        )
        self._ui.pane_tabs.tabBar().setTabButton(
            0, QtWidgets.QTabBar.RightSide, self._empty_widget
        )

    def _on_tab_close_request(self, index: int) -> None:
        self._ui.pane_tabs.removeTab(index)

    def _reload(self) -> None:
        self._ui.outputPane.clear()
        self._ui.errorPane.clear()
        self._load_local_workflow_dock()
        try:
            attributes = self.get_recipe_attributes()
        except RecipeInspectionError:
            # this error has already been posted to the log window
            self.failed_to_load()
            return
        self._update_window_title(attributes)
        self._load_configuration_dock(attributes)
        self._generate_dependency_graph(attributes)

    def _close_recipe(self) -> None:
        sub_window = self.parent()
        sub_window.close()

    def on_local_cache_changed(self, cache_name: str) -> None:
        """
        Slot called when a local cache is changed so that user facing information
        can be updated
        """
        if cache_name != self.recipe.context.cache_name:
            return
        self._ui.behaviourToolbar.refresh_content()

    def _on_configure_packageid_context_menu(self, position: QtCore.QPoint) -> None:
        action = QAction("Copy to clipboard", self)
        action.triggered.connect(self._on_configure_package_id_copy)
        menu = QtWidgets.QMenu(self)
        menu.addAction(action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_configure_package_id_copy(self) -> None:
        QtWidgets.QApplication.clipboard().setText(self._ui.configurePackageId.text())



#!/usr/bin/env python3

"""
Widget utilities
"""

import platform
import typing

from qtpy import QtCore, QtWidgets


def search_for_dir_options() -> QtWidgets.QFileDialog.Options:
    """
    Get standard file dialog options for directory searching.
    """
    options = QtWidgets.QFileDialog.ShowDirsOnly
    if platform.system() == "Darwin":
        # have to use the non-native dialog on macOS, otherwise cannot browse
        # into an application bundle
        options = options | QtWidgets.QFileDialog.DontUseNativeDialog
    return options


def search_for_file_options() -> QtWidgets.QFileDialog.Options:
    """
    Get standard file dialog options for file searching.
    """
    options = QtWidgets.QFileDialog.Options()
    if platform.system() == "Darwin":
        # have to use the non-native dialog on macOS, otherwise cannot browse
        # into an application bundle
        options = options | QtWidgets.QFileDialog.DontUseNativeDialog
    return options


def clear_widgets_from_layout(layout: QtWidgets.QLayout) -> None:
    """
    Clear all widgets from a layout.
    Does not take child layouts into account.
    """
    for i in reversed(range(layout.count())):
        layout.itemAt(i).widget().setParent(None)  # type: ignore


class BlockSignals:
    """
    Use instances of this in with statements, that block and restore signals to the
    provided Qt object in the initialiser.

    These can be nested.

    Returns the object being blocked, for reuse in the scope, as the expression used to
    obtain the object may have been complex.
    """

    def __init__(self, qt_object: QtCore.QObject) -> None:
        self._object = qt_object

    def __enter__(self) -> QtCore.QObject:
        # pylint: disable=attribute-defined-outside-init
        self._old_state = self._object.blockSignals(True)
        return self._object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> None:
        self._object.blockSignals(self._old_state)



#!/usr/bin/env python3

"""
Recipe command log window
"""

import os
import stat
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE6

from cruiz.interop.commandparameters import CommandParameters

if PYSIDE6:
    QAction = QtGui.QAction
else:
    QAction = QtWidgets.QAction


class CommandListWidgetItem(QtWidgets.QListWidgetItem):
    """
    QListWidgetItem representing a Conan command
    """

    def __init__(
        self,
        parameters: CommandParameters,
        parent: typing.Optional[QtWidgets.QListWidget] = None,
    ) -> None:
        if parameters.cwd:
            expression = f"{parameters} (in {parameters.cwd})"
        else:
            expression = f"{parameters}"
        super().__init__(expression, parent, 1000)
        self.setData(0x0100, parameters)


class RecipeCommandHistoryWidget(QtWidgets.QListWidget):
    """
    QListWidget representing a history of Conan commands
    """

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self.customContextMenuRequested.connect(self._show_context_menu)
        self.itemSelectionChanged.connect(self._selection_changed)
        self._export_bash_action = QAction("Export to file...", self)
        self._export_bash_action.triggered.connect(self._export_bash)
        self._export_zsh_action = QAction("Export to file...", self)
        self._export_zsh_action.triggered.connect(self._export_zsh)
        self._export_cmd_action = QAction("Export to file...", self)
        self._export_cmd_action.triggered.connect(self._export_cmd)
        self._copy_bash_to_clipboard_action = QAction("Copy to clipboard", self)
        self._copy_bash_to_clipboard_action.triggered.connect(
            self._copy_bash_to_clipboard
        )
        self._copy_zsh_to_clipboard_action = QAction("Copy to clipboard", self)
        self._copy_zsh_to_clipboard_action.triggered.connect(
            self._copy_zsh_to_clipboard
        )
        self._copy_cmd_to_clipboard_action = QAction("Copy to clipboard", self)
        self._copy_cmd_to_clipboard_action.triggered.connect(
            self._copy_cmd_to_clipboard
        )
        self._menu = QtWidgets.QMenu()
        self._menu.setEnabled(False)
        bash_menu = self._menu.addMenu("bash")
        bash_menu.insertAction(None, self._export_bash_action)  # type: ignore[arg-type]
        bash_menu.insertAction(
            None, self._copy_bash_to_clipboard_action  # type: ignore[arg-type]
        )
        zsh_menu = self._menu.addMenu("zsh")
        zsh_menu.insertAction(None, self._export_zsh_action)  # type: ignore[arg-type]
        zsh_menu.insertAction(
            None, self._copy_zsh_to_clipboard_action  # type: ignore[arg-type]
        )
        cmd_menu = self._menu.addMenu("Batch CMD")
        cmd_menu.insertAction(None, self._export_cmd_action)  # type: ignore[arg-type]
        cmd_menu.insertAction(
            None, self._copy_cmd_to_clipboard_action  # type: ignore[arg-type]
        )

    def _show_context_menu(self, position: QtCore.QPoint) -> None:
        self._menu.exec_(self.mapToGlobal(position))

    def _selection_changed(self) -> None:
        enabled = bool(self.selectedItems())
        self._menu.setEnabled(enabled)

    def _export_bash(self) -> None:
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Export Conan command as bash shell script",
            "",
            "Shell script (*.sh)",
        )
        if not filename:
            return
        item = self.selectedItems()[0]
        parameters = item.data(0x0100)
        with open(filename, "wt", encoding="utf-8") as shell_script:
            shell_script.write("#!/usr/bin/env bash\n")
            shell_script.write(f"{parameters.bash_expression.getvalue()}\n")
        os.chmod(filename, os.stat(filename).st_mode | stat.S_IEXEC)

    def _export_zsh(self) -> None:
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Export Conan command as zsh shell script",
            "",
            "Shell script (*.sh)",
        )
        if not filename:
            return
        item = self.selectedItems()[0]
        parameters = item.data(0x0100)
        with open(filename, "wt", encoding="utf-8") as shell_script:
            shell_script.write("#!/usr/bin/env zsh\n")
            shell_script.write(f"{parameters.zsh_expression.getvalue()}\n")
        os.chmod(filename, os.stat(filename).st_mode | stat.S_IEXEC)

    def _export_cmd(self) -> None:
        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            self,
            "Export Conan command as CMD Batch shell script",
            "",
            "Batch script (*.bat)",
        )
        if not filename:
            return
        item = self.selectedItems()[0]
        parameters = item.data(0x0100)
        with open(filename, "wt", encoding="utf-8") as shell_script:
            shell_script.write(f"{parameters.cmd_expression.getvalue()}\n")
        os.chmod(filename, os.stat(filename).st_mode | stat.S_IEXEC)

    def _copy_bash_to_clipboard(self) -> None:
        item = self.selectedItems()[0]
        parameters = item.data(0x0100)
        qApp.clipboard().setText(  # type: ignore  # noqa: F821
            parameters.bash_expression.getvalue()
        )

    def _copy_zsh_to_clipboard(self) -> None:
        item = self.selectedItems()[0]
        parameters = item.data(0x0100)
        qApp.clipboard().setText(  # type: ignore  # noqa: F821
            parameters.zsh_expression.getvalue()
        )

    def _copy_cmd_to_clipboard(self) -> None:
        item = self.selectedItems()[0]
        parameters = item.data(0x0100)
        qApp.clipboard().setText(  # type: ignore  # noqa: F821
            parameters.cmd_expression.getvalue()
        )



#!/usr/bin/env python3

"""
Conan config entries of interest
"""

from enum import Enum


class ConanConfigBoolean(Enum):
    """
    Enumeration for Conan configuration items that are Boolean values
    """

    PRINT_RUN_COMMANDS = "log.print_run_commands"
    REVISIONS = "general.revisions_enabled"



#!/usr/bin/env python3

"""
Utils for worker context managers
"""

from __future__ import annotations

import datetime
import multiprocessing
import os
import traceback
import typing

from qtpy import QtCore

from cruiz.interop.commandparameters import CommandParameters
from cruiz.interop.commonparameters import CommonParameters
from cruiz.interop.searchrecipesparameters import SearchRecipesParameters
from cruiz.interop.reciperevisionsparameters import RecipeRevisionsParameters
from cruiz.interop.packagerevisionsparameters import PackageRevisionsParameters
from cruiz.interop.packageidparameters import PackageIdParameters
from cruiz.interop.packagebinaryparameters import PackageBinaryParameters

from .env import set_env, clear_conan_env
from .message import Message, Failure, Stdout
from .text2html import text_to_html


class Worker:
    """
    Non-Conan specific worker context manager
    """

    # TODO: tried to use typing.Type[CommonParameters] here but mypy didn't like it
    def __init__(
        self,
        queue: multiprocessing.Queue[Message],
        params: typing.Union[
            CommandParameters,
            SearchRecipesParameters,
            RecipeRevisionsParameters,
            PackageRevisionsParameters,
            PackageIdParameters,
            PackageBinaryParameters,
        ],
    ):
        self._queue = queue
        self._params = params
        if isinstance(params, CommandParameters):
            self._wall_clock = QtCore.QElapsedTimer() if params.time_commands else None
        else:
            # can occur for other types of *Parameters classes
            self._wall_clock = None

    def __enter__(self) -> None:
        multiprocessing.get_logger().debug("%i (child): %s", os.getpid(), self._params)
        clear_conan_env()
        if isinstance(self._params, CommandParameters):
            set_env(self._params.added_environment, self._params.removed_environment)
        elif isinstance(self._params, CommonParameters):
            set_env(self._params.added_environment, self._params.removed_environment)
        else:
            try:
                if "env" in self._params:
                    set_env(self._params["env"], [])
            except TypeError:
                # can happen for other types of *Parameters classes
                pass

        if self._wall_clock is not None:
            self._wall_clock.start()

    def _exception_to_html(
        self, exc_type: typing.Any, value: typing.Any, exc_tb: typing.Any
    ) -> None:
        # pylint: disable=unused-argument

        # since tracebacks objects can't be passed over the multiprocessing queue,
        # encode it into the string of a new Exception object
        # (also removes the issue of Conan exception types passing the divide)
        exc_text = traceback.format_exception(exc_type, value, exc_tb)
        html = "<font color='red'>"
        html += text_to_html("".join(exc_text))
        html += "</font>"

        self._queue.put(Failure(Exception(html)))

        multiprocessing.get_logger().debug(
            "%i (child): Exception detected: %s",
            os.getpid(),
            exc_text,
        )

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        if exc_value:
            self._exception_to_html(exc_type, exc_value, exc_traceback)
        if self._wall_clock is not None:
            elapsed_time = self._wall_clock.elapsed()
            self._queue.put(Stdout("-" * 64))
            if isinstance(self._params, CommandParameters):
                worker = self._params.worker
            else:
                # TODO: does this get used any more?
                worker = self._params["worker"]  # type: ignore
            self._queue.put(
                Stdout(
                    f"Command {worker} ran in "
                    f"{datetime.timedelta(seconds=elapsed_time / 1000)}"
                )
            )
            self._queue.put(Stdout("-" * 64))
        self._queue.close()
        self._queue.join_thread()
        return True  # suppress further exception propogation


# pylint: disable=too-few-public-methods
class ConanWorker(Worker):
    """
    Conan specific context manager
    """

    def __enter__(self) -> typing.Any:
        super().__enter__()
        # import here because it can use the environment variables set in the
        # super class
        # pylint: disable=import-outside-toplevel
        from .conanapi import instance

        api = instance(self._queue, self._params)
        return api


def replace_conan_version_struct_with_string(
    result: typing.Dict[typing.Any, typing.Any]
) -> None:
    """
    Results have ConanVersion structs, but these cannot be passed over the
    process divide so just convert them to strings
    """
    if "installed" not in result:
        return
    for installed in result["installed"]:
        recipe = installed["recipe"]
        recipe["version"] = str(recipe["version"])
        for package in installed["packages"]:
            try:
                cpp_info = package["cpp_info"]
            except KeyError:
                # no cpp_info, so abort further processing
                continue
            cpp_info["version"] = str(cpp_info["version"])
            try:
                for component_key in cpp_info["components"]:
                    component = cpp_info["components"][component_key]
                    component["version"] = str(component["version"])
            except KeyError:
                # ignore 'components' not being in cpp_info
                pass
            try:
                build_modules: typing.Dict[str, str] = {}
                for key, value in cpp_info["build_modules"].items():
                    build_modules[key] = value
                cpp_info["build_modules"] = build_modules
            except KeyError:
                # ignore 'build_modules' not being in cpp_info
                pass



#!/usr/bin/env python3

"""
Build features toolbar
"""

import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

from cruiz.constants import CompilerCacheTypes

from cruiz.settings.managers.recipe import (
    RecipeSettingsReader,
    RecipeSettings,
    RecipeSettingsWriter,
)
from cruiz.settings.managers.compilercachepreferences import CompilerCacheSettingsReader

from cruiz.widgets.util import BlockSignals

if PYSIDE2:
    from cruiz.pyside2.recipe_cmake_features_frame import Ui_cmakeFeaturesFrame
    from cruiz.pyside2.recipe_compilercache_features_frame import (
        Ui_compilerCacheFrame,
    )
    from cruiz.pyside2.recipe_compiler_cache_configuration_dialog import (
        Ui_CompilerCacheConfigurationDialog,
    )
else:
    from cruiz.pyside6.recipe_cmake_features_frame import Ui_cmakeFeaturesFrame
    from cruiz.pyside6.recipe_compilercache_features_frame import (
        Ui_compilerCacheFrame,
    )
    from cruiz.pyside6.recipe_compiler_cache_configuration_dialog import (
        Ui_CompilerCacheConfigurationDialog,
    )


class _CMakeFeaturesFrame(QtWidgets.QFrame):
    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self._ui = Ui_cmakeFeaturesFrame()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._uuid: typing.Optional[QtCore.QUuid] = None
        self._ui.cmakeFindDebug.stateChanged.connect(self._toggle_cmake_find_debug_mode)
        self._ui.cmakeVerbose.stateChanged.connect(self._toggle_cmake_verbose)

    def refresh_content(self, uuid: QtCore.QUuid) -> None:
        """
        Refresh the content of the frame for the specified recipe UUID
        """
        self._uuid = uuid
        with RecipeSettingsReader.from_uuid(uuid) as settings:
            find_debug = settings.cmake_find_debug.resolve()
            verbose = settings.cmake_verbose.resolve()
        with BlockSignals(self._ui.cmakeFindDebug) as blocked_widget:
            blocked_widget.setChecked(find_debug)
        with BlockSignals(self._ui.cmakeVerbose) as blocked_widget:
            blocked_widget.setChecked(verbose)

    def _toggle_cmake_find_debug_mode(self, state: int) -> None:
        is_checked = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        settings = RecipeSettings()
        settings.cmake_find_debug = is_checked  # type: ignore
        assert self._uuid
        RecipeSettingsWriter.from_uuid(self._uuid).sync(settings)

    def _toggle_cmake_verbose(self, state: int) -> None:
        is_checked = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        settings = RecipeSettings()
        settings.cmake_verbose = is_checked  # type: ignore
        assert self._uuid
        RecipeSettingsWriter.from_uuid(self._uuid).sync(settings)


class _CompilerCacheConfigurationDialog(QtWidgets.QDialog):
    modified = QtCore.Signal()

    def __init__(
        self, uuid: QtCore.QUuid, parent: typing.Optional[QtWidgets.QWidget] = None
    ) -> None:
        super().__init__(parent)
        self._ui = Ui_CompilerCacheConfigurationDialog()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._uuid = uuid
        self._settings = RecipeSettings.from_uuid(uuid)
        # -- signals
        self.modified.connect(self._modification_applied)
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply
        ).setEnabled(False)
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).setEnabled(False)
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply
        ).clicked.connect(self._save)
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).clicked.connect(self._save)
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).clicked.connect(self.accept)
        self._ui.ccache_arguments.textChanged.connect(self._change_ccache_arguments)
        self._ui.sccache_arguments.textChanged.connect(self._change_sccache_arguments)
        self._ui.buildcache_arguments.textChanged.connect(
            self._change_buildcache_arguments
        )
        self._load_defaults()

    def _load_defaults(self) -> None:
        with RecipeSettingsReader.from_uuid(self._uuid) as settings:
            args = settings.compilercache_autotools_configuration.resolve()
        if CompilerCacheTypes.CCACHE.name in args:
            with BlockSignals(self._ui.ccache_arguments) as blocked_widget:
                blocked_widget.setText(args[CompilerCacheTypes.CCACHE.name])
        if CompilerCacheTypes.SCCACHE.name in args:
            with BlockSignals(self._ui.sccache_arguments) as blocked_widget:
                blocked_widget.setText(args[CompilerCacheTypes.SCCACHE.name])
        if CompilerCacheTypes.BUILDCACHE.name in args:
            with BlockSignals(self._ui.buildcache_arguments) as blocked_widget:
                blocked_widget.setText(args[CompilerCacheTypes.BUILDCACHE.name])

    def _any_modifications(self) -> bool:
        return not self._settings.empty(RecipeSettingsReader.from_uuid(self._uuid))

    def _modification_applied(self) -> None:
        enabled = self._any_modifications()
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Apply
        ).setEnabled(enabled)
        self._ui.buttonBox.button(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
        ).setEnabled(enabled)

    def _save(self) -> None:
        RecipeSettingsWriter.from_uuid(self._uuid).sync(self._settings)
        self.modified.emit()

    def reject(self) -> None:
        if self._any_modifications():
            response = QtWidgets.QMessageBox.question(
                self,
                "Unsaved compiler cache arguments",
                "Modifications are unsaved. Do you want to discard them?",
            )
            if response == QtWidgets.QMessageBox.No:
                return
        super().reject()

    def _change_ccache_arguments(self, text: str) -> None:
        self._settings.append_compilercache_autotools_configuration(
            CompilerCacheTypes.CCACHE, text or None
        )
        self.modified.emit()

    def _change_sccache_arguments(self, text: str) -> None:
        self._settings.append_compilercache_autotools_configuration(
            CompilerCacheTypes.SCCACHE, text or None
        )
        self.modified.emit()

    def _change_buildcache_arguments(self, text: str) -> None:
        self._settings.append_compilercache_autotools_configuration(
            CompilerCacheTypes.BUILDCACHE, text or None
        )
        self.modified.emit()


class _CompilerCacheFeaturesFrame(QtWidgets.QFrame):
    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self._ui = Ui_compilerCacheFrame()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._uuid: typing.Optional[QtCore.QUuid] = None
        self._ui.useCache.stateChanged.connect(self._toggle_use_cache)
        self._ui.chooseCache.currentTextChanged.connect(self._change_cache)
        self._ui.configureCache.clicked.connect(self._open_configuration)

    def refresh_content(self, uuid: QtCore.QUuid) -> None:
        """
        Refresh the content of the frame for the specified recipe UUID
        """
        self._uuid = uuid
        with RecipeSettingsReader.from_uuid(uuid) as settings:
            compiler_cache = settings.compiler_cache.resolve()
        with BlockSignals(self._ui.useCache) as blocked_widget:
            blocked_widget.setChecked(bool(compiler_cache))
        with BlockSignals(self._ui.chooseCache) as blocked_widget:
            blocked_widget.setEnabled(bool(compiler_cache))
            default_compiler_cache = self._identify_default_compiler_cache()
            blocked_widget.setCurrentText(compiler_cache or default_compiler_cache)

    def _identify_default_compiler_cache(self) -> str:
        with CompilerCacheSettingsReader() as settings:
            default_compiler_cache = settings.default.resolve()
        default_index = self._ui.chooseCache.findText(default_compiler_cache)
        assert default_index >= 0
        for i in range(self._ui.chooseCache.count()):
            self._ui.chooseCache.setItemIcon(
                i,
                QtGui.QIcon(":/cruiz.png") if i == default_index else QtGui.QIcon(),
            )
        return default_compiler_cache

    def _toggle_use_cache(self, state: int) -> None:
        is_checked = state == QtCore.Qt.Checked  # type: ignore[comparison-overlap]
        settings = RecipeSettings()
        cache_name = self._ui.chooseCache.currentText() if is_checked else None
        settings.compiler_cache = cache_name  # type: ignore[assignment]
        assert self._uuid
        RecipeSettingsWriter.from_uuid(self._uuid).sync(settings)
        self._ui.chooseCache.setEnabled(is_checked)
        self.refresh_content(self._uuid)

    def _change_cache(self, text: str) -> None:
        settings = RecipeSettings()
        settings.compiler_cache = text  # type: ignore
        assert self._uuid
        RecipeSettingsWriter.from_uuid(self._uuid).sync(settings)

    def _open_configuration(self) -> None:
        assert self._uuid
        _CompilerCacheConfigurationDialog(self._uuid, self).exec_()


class BuildFeaturesToolbar(QtWidgets.QToolBar):
    """
    QToolBar representing the build features of a recipe, CMake and compiler caching
    """

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self._cmake_features = _CMakeFeaturesFrame()
        self._compilercache_features = _CompilerCacheFeaturesFrame()
        self.addWidget(self._cmake_features)
        self.addSeparator()
        self.addWidget(self._compilercache_features)
        self.addSeparator()

    def refresh_content(self, uuid: QtCore.QUuid) -> None:
        """
        Refresh the content of the toolbar.
        """
        self._cmake_features.refresh_content(uuid)
        self._compilercache_features.refresh_content(uuid)



#!/usr/bin/env python3

"""
Debug utility to recursively identify all the types of all attributes in an object.
Useful to figure out if Conan types made it across the process divide.

Define CRUIZ_DUMP_OBJECT_TYPES_DETAIL in the environment to show more detailed object
information (which is much harder to read).
"""

import logging
import os
import typing

logger = logging.getLogger(__name__)

__VISITED: typing.List[typing.Any] = []


def __examine(obj: typing.Any, loglevel: int, depth: int, prefix: str = "") -> None:
    if id(obj) in __VISITED:
        return
    __VISITED.append(id(obj))
    indent = depth * " "
    if "CRUIZ_DUMP_OBJECT_TYPES_DETAIL" in os.environ:
        logger.log(loglevel, "%s%s%s (%s)", indent, prefix, type(obj), obj.__repr__)
    else:
        logger.log(loglevel, "%s%s%s", indent, prefix, type(obj))
    if isinstance(obj, (list, tuple)):
        for i, item in enumerate(obj):
            __examine(item, loglevel, depth + 1, f"{i}: ")
    elif isinstance(obj, str):
        pass
    elif isinstance(obj, bool):
        pass
    elif isinstance(obj, dict):
        for key, value in obj.items():
            __examine(value, loglevel, depth + 1, f"{key}: ")
    elif obj is not None:
        try:
            dict_of_attribs = obj.__dict__
            for key, value in dict_of_attribs.items():
                __examine(value, loglevel, depth + 1, f"{key}: ")
        except AttributeError:
            logger.log(
                loglevel,
                "%s\tWARNING: Type is opaque - no attribute information",
                indent,
            )


def dump_object_types(obj: typing.Any, loglevel: str = "DEBUG") -> None:
    """
    Recursively dump the types of all attributes of the specified object
    Optional logging level for the output
    """
    __VISITED.clear()
    __examine(obj, int(logging.getLevelName(loglevel)), 0)



#!/usr/bin/env python3

"""
Representation of a recipe
"""

import pathlib
import typing

from qtpy import QtCore

from cruiz.commands.context import ConanContext
from cruiz.commands.logdetails import LogDetails


# TODO: need to remember why this is derived from QObject?
# TODO: possibly due to the parent-child ownership, and thus lifetime
class Recipe(QtCore.QObject):
    """
    Sufficient information to uniquely identify a recipe
    """

    def __init__(
        self,
        path: pathlib.Path,
        uuid: QtCore.QUuid,
        cache_name: str,
        log_details: LogDetails,
        parent: QtCore.QObject,
    ) -> None:
        super().__init__(parent)
        self.path = path
        self.uuid = uuid
        self.context = ConanContext(cache_name, log_details)

    def close(self) -> None:
        """
        Close the recipe
        """
        self.context.close()

    @property
    def version(self) -> typing.Optional[str]:
        """
        Get the version of the recipe.
        May be None.
        """
        from cruiz.settings.managers.recipe import RecipeSettingsReader

        with RecipeSettingsReader.from_uuid(self.uuid) as settings:
            attributes = settings.attribute_overrides.resolve()
        return attributes.get("version")

    @property
    def user(self) -> typing.Optional[str]:
        """
        Get the user namespace of the recipe.
        May be None.
        """
        from cruiz.settings.managers.recipe import RecipeSettingsReader

        with RecipeSettingsReader.from_uuid(self.uuid) as settings:
            attributes = settings.attribute_overrides.resolve()
        return attributes.get("user")

    @property
    def channel(self) -> typing.Optional[str]:
        """
        Get the channel namespace of the recipe.
        May be None.
        """
        from cruiz.settings.managers.recipe import RecipeSettingsReader

        with RecipeSettingsReader.from_uuid(self.uuid) as settings:
            attributes = settings.attribute_overrides.resolve()
        return attributes.get("channel")

    @property
    def folder(self) -> pathlib.Path:
        """
        Get the folder containing the recipe.
        """
        return self.path.parent



#!/usr/bin/env python3

"""
Interop between Conan and cruiz
"""

from __future__ import annotations

from dataclasses import dataclass, field
import typing


@dataclass
class PackageNode:
    """
    Representation of a node in the Conan dependency graph
    """

    name: str
    reference: str
    package_id: str
    recipe_revision: str
    short_paths: bool
    info: typing.Optional[str]
    is_runtime: bool
    children: typing.List[PackageNode] = field(default_factory=list)
    parents: typing.List[PackageNode] = field(default_factory=list)

    def clone_standalone(self) -> PackageNode:
        return PackageNode(
            self.name,
            self.reference,
            self.package_id,
            self.recipe_revision,
            self.short_paths,
            self.info,
            self.is_runtime,
            [],
            [],
        )



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success, Stdout


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan remove [-f] ref'
    """
    with worker.ConanWorker(queue, params) as api:
        assert params.package_reference
        args = {}
        if params.force:
            args["force"] = params.force
        result = api.remove(
            params.package_reference,
            **args,
        )
        queue.put(
            Stdout(
                "Removed packages in the local cache matching pattern "
                f'"{params.package_reference}"'
                " (forced)"
                if params.force
                else ""
            )
        )

        queue.put(Success(result))



#!/usr/bin/env python3

"""
Qt model for recent Conan remotes
"""

import typing

from qtpy import QtCore


class RecentConanRemotesModel(QtCore.QAbstractListModel):
    """
    Model representing recent Conan remotes
    """

    def __init__(self) -> None:
        super().__init__()
        self._urls: typing.Optional[typing.List[str]] = None

    def set_urls(self, urls: typing.List[str]) -> None:
        """
        Set the URLs for the Conan remotes.
        """
        self.beginResetModel()
        self._urls = urls
        self.endResetModel()

    def url(self, index: int) -> str:
        """
        Get a Conan remote identified by its index
        """
        assert self._urls
        return self._urls[index]

    def rowCount(self, parent) -> int:  # type: ignore
        if parent.isValid():
            return 0
        if self._urls is None:
            return 0
        return len(self._urls)

    def headerData(self, section, orientation, role):  # type: ignore
        # pylint: disable=unused-argument
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            return "URL"
        return None

    def data(self, index, role):  # type: ignore
        if role == QtCore.Qt.DisplayRole:
            assert self._urls
            return self._urls[index.row()]
        return None



#!/usr/bin/env python3

"""
Settings context manager for named local caches
"""

from __future__ import annotations

import typing

from cruiz.interop.pod import ExtraProfileDirectory

from .writermixin import _WriterMixin
from .valueclasses import ScalarValue, DictValue, ListValue

from .basesettings import (
    BaseSettings,
    StringSetting,
    DictSetting,
    CommonSettings,
    SettingMeta,
    ListSetting,
)


# TODO: this needs some work as there is a lot of repetition


class NamedLocalCacheSettings(CommonSettings):
    """
    Settings for named local caches.
    """

    def __init__(self) -> None:
        self._property_meta = {
            "home_dir": SettingMeta("Dir", StringSetting, None, ScalarValue),
            "short_home_dir": SettingMeta("ShortDir", StringSetting, None, ScalarValue),
            "environment_added": SettingMeta("Environment", DictSetting, {}, DictValue),
            "environment_removed": SettingMeta(
                "EnvironmentRemoved", ListSetting, [], ListValue
            ),
            "extra_profile_directories": SettingMeta(
                "ExtraProfileDirs", DictSetting, {}, DictValue
            ),
        }

    # TODO: this is a bit weird, should it be a function?
    # Also, current uses of this function only check to see its length
    @property
    def recipe_uuids(self) -> typing.List[str]:
        """
        Return list of UUIDs for the recipes in this cache.
        """
        results: typing.List[str] = []
        settings = BaseSettings.make_settings()
        assert self.settings_reader
        settings.beginGroup("Recipe")
        for uuid in settings.childGroups():
            if (
                settings.value(f"{uuid}/LocalCacheName")
                == self.settings_reader.cache_name
            ):
                results.append(uuid)
        settings.endGroup()
        return results

    @property
    def home_dir(self) -> StringSetting:
        """
        Return the home directory of this cache.
        """
        return self._get_value_via_meta()

    @home_dir.setter
    def home_dir(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def short_home_dir(self) -> StringSetting:
        """
        Return the short home directory of this cache.
        """
        return self._get_value_via_meta()

    @short_home_dir.setter
    def short_home_dir(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def environment_added(self) -> DictSetting:
        """
        Return the added environment dictionary for this cache.
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "Environment"
        env: typing.Dict[str, str] = {}
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            for i in range(count):
                settings.setArrayIndex(i)
                name = str(settings.value("Name"))
                value = str(settings.value("Value"))
                env[name] = value
        return DictSetting(env, {})

    def environment_added_at(self, row_index: int) -> typing.Tuple[str, str]:
        """
        Get the key-value pair at the specified index.
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "Environment"
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            settings.setArrayIndex(row_index)
            name = str(settings.value("Name"))
            value = str(settings.value("Value"))
            return name, value

    @property
    def environment_removed(self) -> ListSetting:
        """
        Return the removed environment keys for this cache.
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "EnvironmentRemoved"
        env_removed: typing.List[str] = []
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            for i in range(count):
                settings.setArrayIndex(i)
                name = str(settings.value("Name"))
                env_removed.append(name)
        return ListSetting(env_removed, [])

    def sync_env_changes(
        self,
        existing_removed: typing.List[str],
        changes: _EnvChangeManagement,
    ) -> None:
        """
        Synchronise the changes environment into the NamedLocalCacheSettings
        ready to be written back to disk.

        The existing 'removed' environment keys are provided here because it's a list
        which must always be provided in full.
        """
        assert changes.has_change
        if changes.add_required or changes.add_clear:
            env_added: typing.Dict[str, typing.Optional[str]] = {}
            if changes.add_required:
                env_added.update(changes.add_required)
            for key in changes.add_clear:
                env_added[key] = None
            # avoids private name mangling
            setattr(  # noqa: B010
                self,
                "__environment_added",
                DictValue(
                    env_added, "environment_added", "Environment", "Name", "Value"
                ),
            )
        if changes.remove_required or changes.remove_clear:
            if changes.remove_required:
                existing_removed.extend(changes.remove_required)
            for key in changes.remove_clear:
                existing_removed.remove(key)
            # avoids private name mangling
            setattr(  # noqa: B010
                self,
                "__environment_removed",
                ListValue(
                    existing_removed,
                    "environment_removed",
                    "EnvironmentRemoved",
                    "Name",
                ),
            )

    @property
    def extra_profile_directories(self) -> DictSetting:
        """
        Return extra profile directories for this cache.
        """
        assert self.settings_reader
        settings = self.settings_reader.settings
        key = "ExtraProfileDirs"
        dirs: typing.Dict[str, str] = {}
        with BaseSettings.ReadArray(key, settings=settings) as (settings, count):
            for i in range(count):
                settings.setArrayIndex(i)
                name = str(settings.value("Name"))
                value = str(settings.value("Path"))
                dirs[name] = value
        return DictSetting(dirs, {})

    def append_extra_profile_directories(
        self, extra_dir: ExtraProfileDirectory
    ) -> None:
        """
        Append new extra profile directories

        self.__extra_profile_directories set at the end of the function,
        but in case this is called several times, read at the start too
        """
        try:
            (
                existing_extra_profile_dirs,
                prop,
                real_key,
            ) = self.__extra_profile_directories  # type: ignore
        except AttributeError:
            existing_extra_profile_dirs = {}
            prop = "extra_profile_directories"
            real_key = "ExtraProfileDirs"
        existing_extra_profile_dirs[extra_dir.name] = extra_dir.directory
        # avoids private name mangling
        setattr(  # noqa: B010
            self,
            "__extra_profile_directories",
            DictValue(existing_extra_profile_dirs, prop, real_key, "Name", "Path"),
        )

    def remove_extra_profile_directories(self, name: str) -> None:
        """
        Append new extra profile directories

        self.__extra_profile_directories set at the end of the function,
        but in case this is called several times, read at the start too
        """
        try:
            (
                existing_extra_profile_dirs,
                prop,
                real_key,
            ) = self.__extra_profile_directories  # type: ignore
        except AttributeError:
            existing_extra_profile_dirs = {}
            prop = "extra_profile_directories"
            real_key = "ExtraProfileDirs"
        existing_extra_profile_dirs[name] = None
        # avoids private name mangling
        setattr(  # noqa: B010
            self,
            "__extra_profile_directories",
            DictValue(existing_extra_profile_dirs, prop, real_key, "Name", "Path"),
        )


class _EnvChangeManagement:
    """
    Manage the changes (CM) in the requested environment
    """

    def __init__(self) -> None:
        self.add_required: typing.Dict[str, str] = {}
        self.add_clear: typing.List[str] = []
        self.remove_required: typing.List[str] = []
        self.remove_clear: typing.List[str] = []

    def key_required(self, key: str, value: str) -> None:
        """
        The environment key-value pair is required to be used.
        If used on a key that is already stored, the value is overwritten.
        """
        self.add_required[key] = value

    def key_not_required(self, key: str) -> None:
        """
        The environment key is no longer required.
        If added in this CM, it is no longer added.
        Otherwise, it is already in the environment and will be scheduled to be removed.
        """
        if key in self.add_required:
            del self.add_required[key]
        else:
            self.add_clear.append(key)

    def key_never_required(self, key: str) -> None:
        """
        The key is scheduled to not be present in the applied environment.
        """
        self.remove_required.append(key)

    def key_not_never_required(self, key: str) -> None:
        """
        If the key was scheduled in this CM not to be present, this is undone.
        Otherwise, the key will no longer be removed.
        """
        if key in self.remove_required:
            self.remove_required.remove(key)
        else:
            self.remove_clear.append(key)

    @property
    def has_change(self) -> bool:
        """
        Determine whether any change management is required.
        """
        return (
            bool(self.add_required)
            or bool(self.add_clear)
            or bool(self.remove_required)
            or bool(self.remove_clear)
        )


class NamedLocalCacheSettingsReader:
    """
    Context manager reader for named local cache settings.
    """

    def __init__(self, cache_name: str) -> None:
        self.cache_name = cache_name
        self.group = BaseSettings._location_of_key_value_pair(
            "Conan/LocalCaches", "Name", cache_name
        )
        self.settings = BaseSettings.make_settings()

    def __enter__(self) -> NamedLocalCacheSettings:
        self.settings.beginGroup(self.group)
        self._settings_object = NamedLocalCacheSettings()
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endGroup()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class NamedLocalCacheSettingsWriter(_WriterMixin):
    """
    Context manager to write named local cache settings to disk.
    """

    def __init__(self, name: str) -> None:
        self._reader_for_writer = NamedLocalCacheSettingsReader(name)


class AllNamedLocalCacheSettingsReader:
    """
    Context manager for returning the names of all local caches.
    """

    def __init__(self) -> None:
        self._settings = BaseSettings.make_settings()

    def __enter__(self) -> typing.List[str]:
        all_names: typing.List[str] = []
        size = self._settings.beginReadArray("Conan/LocalCaches")
        for i in range(size):
            self._settings.setArrayIndex(i)
            all_names.append(str(self._settings.value("Name")))
        self._settings.endArray()
        return all_names

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        return False


class NamedLocalCacheDeleter:
    """
    Delete a named local cache.
    """

    def __init__(self) -> None:
        self._group = "Conan"
        self._settings = BaseSettings.make_settings()

    def delete(self, name: str) -> None:
        """
        Delete the named local cache from settings.
        """
        caches: typing.List[typing.Dict[str, object]] = []
        with BaseSettings.ReadArray(f"{self._group}/LocalCaches", self._settings) as (
            settings,
            count,
        ):
            for i in range(count):
                settings.setArrayIndex(i)
                if settings.value("Name") == name:
                    continue
                keys = settings.childKeys()
                cache = {}
                for key in keys:
                    cache[key] = settings.value(key)
                caches.append(cache)
        assert len(caches) >= 1  # always at least one because of the Default
        with BaseSettings.WriteArray(
            f"{self._group}/LocalCaches", replace=True, settings=self._settings
        ) as settings:
            for i, cache_data in enumerate(caches):
                settings.setArrayIndex(i)
                for key, value in cache_data.items():
                    settings.setValue(key, value)


class NamedLocalCacheCreator:
    """
    Create a named local cache.
    """

    def __init__(self) -> None:
        self._group = "Conan/LocalCaches"
        self._settings = BaseSettings.make_settings()

    def create(
        self,
        name: str,
        home_dir: typing.Optional[str],
        short_home_dir: typing.Optional[str],
    ) -> None:
        """
        Create the named local cache from settings.
        """
        with BaseSettings.ReadArray(self._group, self._settings) as (_, count):
            pass
        with BaseSettings.WriteArray(self._group, settings=self._settings) as settings:
            settings.setArrayIndex(count)
            settings.setValue("Name", name)
            settings.setValue("Dir", home_dir or None)
            settings.setValue("ShortDir", short_home_dir or None)



#!/usr/bin/env python3

"""
Util module to convert colorama escpe codes to HTML
"""

import copy
from io import StringIO
import re

import colorama

from .text2html import text_to_html


def convert_from_colorama_to_html(escaped_string: str) -> str:
    """
    Attempt to convert colorama escape sequences into something that HTML can use
    """
    # pylint: disable=anomalous-backslash-in-string,too-many-branches
    if escaped_string.endswith("\n"):
        # last new line is catered for by the HTML block
        escaped_string = escaped_string.rsplit("\n", 1)[0]
    expr = "(\\x1b\[(\d)+m)"  # ESC[NUMm - find NUM  # noqa: W605
    matches = re.findall(expr, escaped_string)
    non_escaped_string = copy.deepcopy(escaped_string)
    style = None
    foreground_colour = None
    for match in matches:
        code = match[0]
        if code == colorama.Style.BRIGHT:
            style = ("<strong>", "</strong>")
        elif code == colorama.Fore.BLACK:
            foreground_colour = ('<font color="black">', "</font>")
        elif code == colorama.Fore.RED:
            foreground_colour = ('<font color="red">', "</font>")
        elif code == colorama.Fore.GREEN:
            foreground_colour = ('<font color="green">', "</font>")
        elif code == colorama.Fore.YELLOW:
            foreground_colour = ('<font color="yellow">', "</font>")
        elif code == colorama.Fore.BLUE:
            foreground_colour = ('<font color="blue">', "</font>")
        elif code == colorama.Fore.MAGENTA:
            foreground_colour = ('<font color="magenta">', "</font>")
        elif code == colorama.Fore.CYAN:
            foreground_colour = ('<font color="cyan">', "</font>")
        elif code == colorama.Fore.WHITE:
            foreground_colour = ('<font color="white">', "</font>")
        else:
            pass
        non_escaped_string = non_escaped_string.replace(match[0], "")

    html = StringIO()
    if foreground_colour:
        html.write(foreground_colour[0])
    if style:
        html.write(style[0])
    html.write(text_to_html(non_escaped_string))
    if style:
        html.write(style[1])
    if foreground_colour:
        html.write(foreground_colour[1])
    return html.getvalue()



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from .utils.message import Message, End


def invoke(queue: multiprocessing.Queue[Message]) -> None:
    """
    Run an arbitrary command
    """
    queue.put(End())
    queue.close()
    queue.join_thread()



#!/usr/bin/env python3

"""
Remote browser
"""

import logging
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

if PYSIDE2:
    from cruiz.pyside2.remote_browser import Ui_remotebrowser

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.remote_browser import Ui_remotebrowser

    QAction = QtGui.QAction

from cruiz.commands.context import ConanContext
from cruiz.commands.logdetails import LogDetails
from cruiz.constants import DEFAULT_CACHE_NAME

logger = logging.getLogger(__name__)


class RemoteBrowserDock(QtWidgets.QDockWidget):
    """
    Dock widget for browsing Conan remotes
    """

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        logger.debug("+=%d", id(self))
        super().__init__(parent)
        self.setVisible(False)
        self._ui = Ui_remotebrowser()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._ui.log.hide()
        self._ui.log.customContextMenuRequested.connect(self._log_context_menu)
        self._log_details = LogDetails(self._ui.log, None, True, False, None)
        self._log_details.logging.connect(self._ui.log.show)
        self._context = ConanContext(DEFAULT_CACHE_NAME, self._log_details)
        self._ui.pkgref.setup(self._ui)
        self._ui.rrev.setup(self._ui)
        self._ui.package_id.setup(self._ui)
        self._ui.prev.setup(self._ui)
        self._ui.pbinary.setup(self._ui)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        # although this looks like a no-op, it regenerates necessary resources
        self._context.change_cache(self._context.cache_name)
        super().showEvent(event)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if self._context.is_busy:
            QtWidgets.QMessageBox.warning(
                self,
                "Remote browser cannot be closed",
                "Commands are still running. Cannot close remote browser.",
            )
            event.ignore()
            return
        self._context.close()
        super().closeEvent(event)

    def cleanup(self) -> None:
        self._ui.stackedWidget.currentWidget().on_cancel()
        self._log_details.stop()
        self._context.close()

    def on_local_cache_modified(self, cache_name: str) -> None:
        """
        Called when a local cache has been changed upstream
        """
        self._ui.pkgref.refresh_local_cache_names()
        if cache_name == self._context.cache_name:
            self._ui.pkgref.on_local_cache_change(cache_name)
            self._ui.stackedWidget.setCurrentWidget(self._ui.pkgref)
            self._clear_log()

    def _log_context_menu(self, position: QtCore.QPoint) -> None:
        menu = self.sender().createStandardContextMenu(position)
        menu.addSeparator()
        clear_action = QAction("Clear", self)
        clear_action.triggered.connect(self._clear_log)
        menu.addAction(clear_action)
        menu.exec_(self.sender().viewport().mapToGlobal(position))

    def _clear_log(self) -> None:
        self._ui.log.clear()
        self._ui.log.hide()



#!/usr/bin/env python3

"""
The Qt application
"""

from __future__ import annotations

import typing

import qdarkstyle

from qtpy import QtCore, QtGui, QtWidgets

from cruiz.settings.ensuredefaultlocalcache import ensure_default_local_cache
from cruiz.settings.managers.fontpreferences import FontSettingsReader, FontUsage
from cruiz.settings.managers.generalpreferences import GeneralSettingsReader


class Application(QtWidgets.QApplication):
    """
    The application
    """

    def __init__(self, argv: typing.List[str]) -> None:
        super().__init__(argv)
        ensure_default_local_cache()
        self.setAttribute(QtCore.Qt.AA_DontUseNativeMenuBar)
        self.setWindowIcon(QtGui.QPixmap(":/cruiz.png"))
        self.default_font = QtWidgets.QApplication.font()
        self.on_preferences_updated()

    @staticmethod
    def instance() -> Application:
        """
        Get the application instance.
        qApp does the job already, but this returns it as an Application so that pylint
        can do static analysis on the additional methods on the subclass
        """
        # pylint: disable=undefined-variable
        return qApp  # type: ignore  # noqa: F821

    def on_preferences_updated(self) -> None:
        """
        Slot executed when preferences have reported an update.
        """
        with GeneralSettingsReader() as settings:
            use_dark_mode = settings.use_dark_mode.resolve()
        self.setStyleSheet(
            qdarkstyle.load_stylesheet_pyside2() if use_dark_mode else ""
        )
        with FontSettingsReader(FontUsage.UI) as settings:
            font_details = (settings.name.resolve(), settings.size.resolve())
        if font_details[0]:
            self.setFont(QtGui.QFont(*font_details))
        else:
            self.setFont(self.default_font)



#!/usr/bin/env python3

"""
Settings context managers for font preferences
"""

from enum import Enum
import typing

from .basesettings import (
    BaseSettings,
    StringSetting,
    IntSetting,
    ComparableCommonSettings,
    SettingMeta,
)
from .writermixin import _WriterMixin
from .valueclasses import ScalarValue


class FontUsage(Enum):
    """
    Each font usage intended
    """

    UI = "UIFont"
    OUTPUT = "OutputPaneFont"


class FontSettings(ComparableCommonSettings):
    """
    Representation of font settings
    """

    def __init__(self) -> None:
        self._property_meta = {
            "name": SettingMeta("FontName", StringSetting, None, ScalarValue),
            "size": SettingMeta("FontSize", IntSetting, None, ScalarValue),
        }

    @property
    def name(self) -> StringSetting:
        """
        Get the name of the font
        """
        return self._get_value_via_meta()

    @name.setter
    def name(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def size(self) -> IntSetting:
        """
        Get the size of the font
        """
        return self._get_value_via_meta()

    @size.setter
    def size(self, value: int) -> None:
        self._set_value_via_meta(value)


class FontSettingsReader:
    """
    Context manager to read font settings from disk
    """

    def __init__(self, usage: FontUsage) -> None:
        self.settings = BaseSettings.make_settings()
        self.group = usage.value

    def __enter__(self) -> FontSettings:
        self.settings.beginGroup(self.group)
        self._settings_object = FontSettings()
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endGroup()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class FontSettingsWriter(_WriterMixin):
    """
    Utility class for writing font settings to disk
    """

    def __init__(self, usage: FontUsage) -> None:
        self._reader_for_writer = FontSettingsReader(usage)



#!/usr/bin/env python3

"""
Settings base classes
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
import inspect
import typing

from qtpy import QtCore, QtGui


@dataclass
class SettingMeta:
    settings_key: str
    type: typing.Type[typing.Any]
    default_value: typing.Any
    save_type: typing.Type[typing.Any]


@dataclass
class StringSetting:
    """
    Representation of a string value from disk settings and its fallback
    """

    value: typing.Optional[str]
    fallback: str

    def resolve(self) -> str:
        """
        If a value exists on disk, return that, otherwise return the fallback.
        """
        if self.value is None:
            return self.fallback
        return self.value


@dataclass
class IntSetting:
    """
    Representation of an integer value from disk settings and its fallback
    """

    value: typing.Optional[int]
    fallback: int

    def resolve(self) -> int:
        """
        If a value exists on disk, return that, otherwise return the fallback.
        """
        if self.value is None:
            return self.fallback
        return self.value


@dataclass
class BoolSetting:
    """
    Representation of a Boolean value from disk settings and its fallback
    """

    value: typing.Optional[bool]
    fallback: bool

    def resolve(self) -> bool:
        """
        If a value exists on disk, return that, otherwise return the fallback.
        """
        if self.value is None:
            return self.fallback
        return self.value


@dataclass
class ColourSetting:
    """
    Representation of a QColor value from disk settings and its fallback
    """

    value: typing.Optional[QtGui.QColor]
    fallback: QtGui.QColor

    def resolve(self) -> QtGui.QColor:
        """
        If a value exists on disk, return that, otherwise return the fallback.
        """
        if self.value is None:
            return self.fallback
        return self.value


@dataclass
class DictSetting:
    """
    Representation of a dictionary from disk settings and its fallback
    """

    value: typing.Optional[typing.Dict[str, str]]
    fallback: typing.Dict[str, str]

    def resolve(self) -> typing.Dict[str, str]:
        """
        If a value exists on disk, return that, otherwise return the fallback.
        """
        if self.value is None:
            return self.fallback
        return self.value

    def __getitem__(self, name: str) -> str:
        assert self.value
        return self.value[name]

    def __setitem__(self, name: str, value: str) -> None:
        assert self.value
        self.value[name] = value

    def __delitem__(self, name: str) -> None:
        assert self.value
        del self.value[name]


@dataclass
class ListSetting:
    """
    Representation of a list from disk settings and its fallback
    """

    value: typing.Optional[typing.List[str]]
    fallback: typing.List[str]

    def resolve(self) -> typing.List[str]:
        """
        If a value exists on disk, return that, otherwise return the fallback.
        """
        if self.value is None:
            return self.fallback
        return self.value


# TODO: does this need renaming?
class WorkflowCwd(IntEnum):
    """
    Supported current working directories in workflow
    """

    RELATIVE_TO_RECIPE = 0
    RELATIVE_TO_GIT = 1


@dataclass
class WorkflowCwdSetting:
    """
    Setting value representing the enum WorkflowCwd, with a fallback
    """

    value: typing.Optional[WorkflowCwd]
    fallback: WorkflowCwd

    def resolve(self) -> WorkflowCwd:
        """
        Resolve the setting value, either the current value or the fallback
        """
        return self.value or self.fallback


class BaseSettings:
    """
    Common settings functionality
    """

    @staticmethod
    def make_settings() -> QtCore.QSettings:
        """
        Make a QSettings instance to operate on settings.
        """
        # since there is a central location to where the QSettings
        # is constructed, override the meanings of the argument to make
        # a sensible path to the preferences file
        return QtCore.QSettings(
            QtCore.QSettings.IniFormat,
            QtCore.QSettings.UserScope,
            "cruiz",  # org name
            "preferences",  # application name
        )

    @staticmethod
    def _location_of_key_value_pair(array_key: str, key: str, value: str) -> str:
        with BaseSettings.ReadArray(array_key) as (settings, size):
            for i in range(size):
                settings.setArrayIndex(i)
                if settings.value(key) == value:
                    return f"{array_key}/{i + 1}"  # settings indices are 1 based
        raise RuntimeError(f"Unable to locate '{key}={value}' in array '{array_key}'")

    class Group:
        """
        Representation of a QSettings group as a context manager
        """

        def __init__(
            self, group: str, settings: typing.Optional[QtCore.QSettings] = None
        ) -> None:
            self._settings = settings or BaseSettings.make_settings()
            self._group = group

        def __enter__(self) -> QtCore.QSettings:
            self._settings.beginGroup(self._group)
            return self._settings

        def __exit__(
            self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
        ) -> typing.Any:
            self._settings.endGroup()

    class ReadArray:
        """
        Representation of a QSettings array as a context manager
        """

        def __init__(
            self, array: str, settings: typing.Optional[QtCore.QSettings] = None
        ) -> None:
            self._settings = settings or BaseSettings.make_settings()
            self._array = array

        def __enter__(self) -> typing.Tuple[QtCore.QSettings, int]:
            count = self._settings.beginReadArray(self._array)
            return self._settings, count

        def __exit__(
            self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
        ) -> typing.Any:
            self._settings.endArray()

    # TODO: this is not such an elegant solution when the array to write
    # is empty and replace=True, because a zero-length array is written to the settings,
    # rather than nothing there is insufficient knowledge here to know whether to start
    # writing the array also, beginWriteArray takes a size parameter too, but doesn't
    # here...
    # it's not an exception either to have a zero-length array, so how is it reported
    # back?
    class WriteArray:
        """
        Representation of writing an array to QSettings as a context manager
        """

        def __init__(
            self,
            array: str,
            replace: bool = False,
            settings: typing.Optional[QtCore.QSettings] = None,
        ) -> None:
            self._settings = settings or BaseSettings.make_settings()
            self._array = array
            self._replace = replace

        def __enter__(self) -> QtCore.QSettings:
            if self._replace:
                self._settings.remove(self._array)
            self._settings.beginWriteArray(self._array)
            return self._settings

        def __exit__(
            self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
        ) -> typing.Any:
            self._settings.endArray()


class CommonSettings:
    """
    Common functionality to all settings classes.

    self._property_meta is expected to be set up in the concrete class
    """

    def __init__(self) -> None:
        self.settings_reader: typing.Optional[typing.Any] = None

    def _get_value_via_meta(self) -> typing.Any:
        """
        Get the value from settings using meta data to determine
        types and settings key
        """
        assert self.settings_reader
        property_name = inspect.getouterframes(inspect.currentframe())[1][3]
        map = self._property_meta[property_name]  # type: ignore
        settings = self.settings_reader.settings
        value = None
        if settings.contains(map.settings_key):
            if map.type == BoolSetting:
                value = settings.value(map.settings_key, type=bool)
            elif map.type in (IntSetting, WorkflowCwdSetting):
                value = settings.value(map.settings_key, type=int)
            else:
                value = settings.value(map.settings_key)
        return map.type(value, map.default_value)

    def _set_value_via_meta(self, value: typing.Any) -> None:
        """
        Set the value into temporary storage before serialising, using
        meta data to know where to read and what to write.
        """
        property_name = inspect.getouterframes(inspect.currentframe())[1][3]
        map = self._property_meta[property_name]  # type: ignore
        value_to_store = map.save_type(value, property_name, map.settings_key)
        attribute_to_store = f"__{property_name}"
        setattr(self, attribute_to_store, value_to_store)


class ComparableCommonSettings(CommonSettings):
    """
    Extended common functionality to settings classes that want to compare their
    current state with that on disk
    """

    def empty(self, reader_context_manager: typing.Any) -> bool:
        """
        Are current settings values different from those on disk?
        """
        keys_to_set = [k for k, _ in self.__dict__.items() if k.startswith("__")]
        if not keys_to_set:
            return True
        # TODO: this obviously has some side-effects, but probably ok
        recheck = False
        with reader_context_manager as settings:
            for key in keys_to_set:
                entry = getattr(self, key)
                value = getattr(settings, entry.property_name)
                # TODO: should this check match the 3-step check in the writer sync()?
                if value.resolve() == entry.value:
                    del self.__dict__[key]
                    recheck = True
        if recheck:
            keys_to_set = [k for k, _ in self.__dict__.items() if k.startswith("__")]
        return not bool(keys_to_set)



#!/usr/bin/env python3

"""
Patching settings
"""

from __future__ import annotations

import pathlib
import shutil
import time
import typing

from qtpy import QtCore

from .managers.basesettings import BaseSettings


CURRENT_SETTINGS_VERSION = 8


class SettingsGroup:
    """
    Context manager for using groups in QSettings
    """

    def __init__(self, settings: QtCore.QSettings, group: str) -> None:
        self._settings = settings
        self._group_name = group

    def __enter__(self) -> None:
        self._settings.beginGroup(self._group_name)

    def __exit__(
        self, exc_type: typing.Any, value: typing.Any, exc_traceback: typing.Any
    ) -> None:
        self._settings.endGroup()


class SettingsReadArray:
    """
    Context manager for reading arrays in QSettings
    """

    def __init__(self, settings: QtCore.QSettings, array: str) -> None:
        self._settings = settings
        self._array_name = array
        self._array_size = 0

    # TODO: why isn't this just a public attribute?
    @property
    def size(self) -> int:
        """
        Property containing the size of the array found in the settings
        """
        return self._array_size

    def __enter__(self) -> SettingsReadArray:
        self._array_size = self._settings.beginReadArray(self._array_name)
        return self

    def __exit__(
        self, exc_type: typing.Any, value: typing.Any, exc_traceback: typing.Any
    ) -> None:
        self._settings.endArray()


class SettingsWriteArray:
    """
    Context manager for writing arrays in QSettings
    """

    def __init__(self, settings: QtCore.QSettings, array: str, size: int) -> None:
        self._settings = settings
        self._array_name = array
        self._array_size = size

    def __enter__(self) -> None:
        self._settings.beginWriteArray(self._array_name, self._array_size)

    def __exit__(
        self, exc_type: typing.Any, value: typing.Any, exc_traceback: typing.Any
    ) -> None:
        self._settings.endArray()


def _patch_settings_from_v7(settings: QtCore.QSettings) -> None:
    """
    Remove
      - NewLoadingBehaviour
      - UseCompactLook
    Rename
      - DefaultRenderDirectory to DefaultRecipeDirectory
    """
    settings.remove("NewLoadingBehaviour")
    settings.remove("UseCompactLook")
    old_defaultrecipedir = settings.value("DefaultRenderDirectory")
    settings.remove("DefaultRenderDirectory")
    if old_defaultrecipedir:
        settings.setValue("DefaultRecipeDirectory", old_defaultrecipedir)


def _patch_settings_from_v6(settings: QtCore.QSettings) -> None:
    """
    V2Mode and UseRevisions move from global settings to per-cache
    If either were true, this was global, so apply to all caches
    """
    old_v2mode = settings.value("Conan/V2Mode", False, type=bool)
    settings.remove("Conan/V2Mode")
    old_userevisions = settings.value("Conan/Revisions", False, type=bool)
    settings.remove("Conan/Revisions")
    if old_v2mode or old_userevisions:
        with SettingsReadArray(settings, "Conan/LocalCaches") as array:
            num_caches = array.size
        with SettingsWriteArray(settings, "Conan/LocalCaches", num_caches):
            for i in range(num_caches):
                settings.setArrayIndex(i)
                if old_v2mode:
                    settings.setValue("V2Mode", old_v2mode)
                if old_userevisions:
                    settings.setValue("Revisions", old_userevisions)


def _patch_settings_from_v5(settings: QtCore.QSettings) -> None:
    """
    Moves recent recipe paths from the recent recipe array to the recipe array.
    """
    recipes = []
    with SettingsReadArray(settings, "RecentRecipes") as array:
        for i in range(array.size):
            settings.setArrayIndex(i)
            recipes.append(
                (
                    settings.value("UUID"),
                    settings.value("Path"),
                )
            )
    settings.remove("RecentRecipes")
    with SettingsWriteArray(settings, "RecentRecipes", len(recipes)):
        for i, (uuid, _) in enumerate(recipes):
            settings.setArrayIndex(i)
            settings.setValue("UUID", uuid)
    for uuid, path in recipes:
        group = f"Recipe/{uuid}"
        with SettingsGroup(settings, group):
            settings.setValue("Path", path)


def _patch_settings_from_v4(settings: QtCore.QSettings) -> None:
    """
    This patch refactors any editables keys.
    "Name" is converted to "Reference", and reformats the value so it's back to being
    a Conan Reference (so matches that in the local cache's editables.json keys) rather
    than a file-system safe component.
    Also the "Path" key is renamed to "RecipePath".
    """
    uuids = []
    with SettingsReadArray(settings, "RecentRecipes") as array:
        for i in range(array.size):
            settings.setArrayIndex(i)
            uuids.append(settings.value("UUID"))
    for uuid in uuids:
        group = f"Recipe/{uuid}"
        with SettingsGroup(settings, group):
            editables: typing.List[typing.Dict[str, object]] = []
            with SettingsReadArray(settings, "Editables") as array:
                for i in range(array.size):
                    editable_data: typing.Dict[str, object] = {}
                    settings.setArrayIndex(i)
                    for key in settings.childKeys():
                        editable_data[key] = settings.value(key)
                    editables.append(editable_data)
            if editables:
                for editable in editables:
                    if "Name" in editable:
                        assert isinstance(editable["Name"], str)
                        editable["Reference"] = editable["Name"].replace("-", "/")
                        del editable["Name"]
                    if "Path" in editable:
                        # this is wrong for managed SCM editables, as the path stored
                        # is relative and there's no way to convert it to an absolute
                        # path as now needed this bits that will go wrong is any
                        # workspace related action in manage editables
                        editable["RecipePath"] = editable["Path"]
                        del editable["Path"]
                settings.remove("Editables")
                with SettingsWriteArray(settings, "Editables", len(editables)):
                    for i, editable in enumerate(editables):
                        settings.setArrayIndex(i)
                        for key, value in editable.items():
                            settings.setValue(key, value)


def _patch_settings_from_v3(settings: QtCore.QSettings) -> None:
    """
    This patch adds a name key to each extra profiles dir for a local cache.
    """
    settings = BaseSettings.make_settings()
    with SettingsReadArray(settings, "Conan/LocalCaches") as cache_array:
        for i in range(cache_array.size):
            settings.setArrayIndex(i)
            with SettingsReadArray(settings, "ExtraProfileDirs") as profiledirs_array:
                extra_dirs = []
                for j in range(profiledirs_array.size):
                    settings.setArrayIndex(j)
                    extra_dirs.append(settings.value("Path"))
            if profiledirs_array.size > 0:
                assert len(extra_dirs) == 1, "Only upgrading a single extra directory"
                settings.remove("ExtraProfileDirs")
                with SettingsWriteArray(
                    settings,
                    "ExtraProfileDirs",
                    len(extra_dirs),
                ):
                    # if there was an extra directory, it was for Nuke,
                    # so name it as such
                    settings.setArrayIndex(0)
                    settings.setValue(
                        "Path",
                        extra_dirs,
                    )
                    settings.setValue("Name", "Nuke")


def _patch_settings_from_v2(settings: QtCore.QSettings) -> None:
    """
    This patch moves Conan V2 from general to conan settings
    """
    old_key = "ConanV2Mode"
    oldv2mode = settings.value(old_key, None)
    if oldv2mode is not None:
        settings.remove(old_key)
        settings.setValue("Conan/V2Mode", oldv2mode)


def _patch_settings_from_v1(settings: QtCore.QSettings) -> None:
    """
    This patch moves editables from a separate group to under the recipe.
    """
    editable_groups = [k for k in settings.childGroups() if k.endswith("-editables")]
    editable_data: typing.Dict[str, typing.List[typing.Dict[str, typing.Any]]] = {}
    for editable_group in editable_groups:
        uuid_str = str(editable_group).replace("-editables", "")
        editable_data[uuid_str] = []
        with SettingsReadArray(settings, editable_group) as array:
            for i in range(array.size):
                settings.setArrayIndex(i)
                per_editable_settings = {}
                keys = settings.allKeys()
                for key in keys:
                    per_editable_settings[key] = settings.value(key)
                editable_data[uuid_str].append(per_editable_settings)
        settings.remove(editable_group)
    for editable_uuid_str, editable_array in editable_data.items():
        group = f"Recipe/{editable_uuid_str}"
        with SettingsGroup(settings, group):
            with SettingsWriteArray(settings, "Editables", len(editable_array)):
                for i, editable_dict in enumerate(editable_array):
                    settings.setArrayIndex(i)
                    for key, value in editable_dict.items():
                        settings.setValue(key, value)


def _patch_settings_from_v0(settings: QtCore.QSettings) -> None:
    # pylint: disable=too-many-locals, too-many-branches
    """
    This patch:
    1) Removes the %General group
    2) converts the QVariant versions of QUuids into strings.
       This applies to values in recent recipes, and keys in recipe settings,
       and keys in editables.
    """
    # move the old general group out into the un-grouped General
    with SettingsGroup(settings, "General"):
        general_settings = {}
        for key in settings.allKeys():
            general_settings[key] = settings.value(key)
    settings.remove("General")
    for key, value in general_settings.items():
        settings.setValue(key, value)
    # recent recipe uuids
    recent_recipes = []
    with SettingsReadArray(settings, "RecentRecipes") as array:
        recent_recipes_size = array.size
        for i in range(recent_recipes_size):
            settings.setArrayIndex(i)
            path = settings.value("Path")
            uuid = settings.value("UUID")
            recent_recipes.append((path, uuid))
    settings.remove("RecentRecipes")
    if recent_recipes_size > 0:
        with SettingsWriteArray(settings, "RecentRecipes", recent_recipes_size):
            for i in range(recent_recipes_size):
                settings.setArrayIndex(i)
                settings.setValue("Path", recent_recipes[i][0])
                assert isinstance(recent_recipes[i][1], QtCore.QUuid)
                settings.setValue(
                    "UUID",
                    recent_recipes[i][1].toString(  # type: ignore
                        QtCore.QUuid.WithoutBraces
                    ),
                )
    # recipe settings
    recipe_settings: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
    with SettingsGroup(settings, "Recipe"):
        uuids = set(settings.childGroups())
        prefix = "QtCore.QUuid('{"
        suffix = "}')"
        for uuid in uuids:
            uuid_str = uuid[len(prefix) : -len(suffix)]  # noqa: E203
            recipe_settings[uuid_str] = {}
            with SettingsGroup(settings, uuid):
                keys = settings.allKeys()
                for key in keys:
                    recipe_settings[uuid_str][key] = settings.value(key)
    settings.remove("Recipe")
    with SettingsGroup(settings, "Recipe"):
        for uuid_str, uuid_dict in recipe_settings.items():
            for key, value in uuid_dict.items():
                settings.setValue(f"{uuid_str}/{key}", value)
    # editables
    suffix = "}')-editables"
    editable_settings: typing.Dict[str, typing.List[typing.Dict[str, typing.Any]]] = {}
    editable_groups = [
        group for group in settings.childGroups() if group.endswith("-editables")
    ]
    for editable in editable_groups:
        uuid_str = editable[len(prefix) : -len(suffix)]  # noqa: E203
        editable_settings[uuid_str] = []
        with SettingsReadArray(settings, editable) as array:
            for i in range(array.size):
                settings.setArrayIndex(i)
                per_editable_settings = {}
                keys = settings.allKeys()
                for key in keys:
                    per_editable_settings[key] = settings.value(key)
                editable_settings[uuid_str].append(per_editable_settings)
        settings.remove(editable)
    for editable_uuid, editable_data in editable_settings.items():
        with SettingsWriteArray(
            settings, f"{editable_uuid}-editables", len(editable_data)
        ):
            for i, data in enumerate(editable_data):
                settings.setArrayIndex(i)
                for key, value in data.items():
                    settings.setValue(key, value)


def _backup_settings(settings: QtCore.QSettings) -> None:
    filename = pathlib.Path(settings.fileName())
    backup_filename = f"{filename.name}.{time.strftime('%Y%m%d-%H%M%S')}.bk"
    backup_path = filename.parent / backup_filename
    shutil.copyfile(filename, backup_path)


def update_settings_to_current_version() -> None:
    """
    Upgrade settings to the latest version, incrementally
    """
    settings = BaseSettings.make_settings()
    if len(settings.allKeys()) > 0:
        version = settings.value("SettingsVersion", 0, type=int)
        assert isinstance(version, int)
        if version < CURRENT_SETTINGS_VERSION:
            _backup_settings(settings)
            while version < CURRENT_SETTINGS_VERSION:
                patch_function = globals()[f"_patch_settings_from_v{version}"]
                patch_function(settings)
                version += 1
    settings.setValue("SettingsVersion", CURRENT_SETTINGS_VERSION)



#!/usr/bin/env python3

"""
Get environment for Conan
"""

import logging
import typing

from cruiz.settings.managers.conanpreferences import ConanSettingsReader
from cruiz.settings.managers.generalpreferences import GeneralSettingsReader
from cruiz.settings.managers.namedlocalcache import NamedLocalCacheSettingsReader


logger = logging.getLogger(__name__)


def get_conan_env(
    cache_name: str,
) -> typing.Tuple[typing.Dict[str, str], typing.List[str]]:
    """
    Get a tuple of
    - a dictionary of environment variables to add
    - a list of environment variable keys to remove
    from the system environment map, in order to invoke Conan commands with
    """
    with NamedLocalCacheSettingsReader(cache_name) as settings:
        home_dir = settings.home_dir.resolve()
        short_home_dir = settings.short_home_dir.resolve()
        added_environment = settings.environment_added.resolve()
        removed_environment = settings.environment_removed.resolve()
    env: typing.Dict[str, str] = {}
    if home_dir:
        env["CONAN_USER_HOME"] = home_dir
    if short_home_dir:
        env["CONAN_USER_HOME_SHORT"] = short_home_dir
    with GeneralSettingsReader() as settings:
        use_dark_mode = settings.use_dark_mode.resolve()
    env["CONAN_COLOR_DARK"] = "0" if use_dark_mode else "1"
    with ConanSettingsReader() as settings:
        log_level = settings.log_level.resolve()
    env["CONAN_LOGGING_LEVEL"] = log_level
    env["CONAN_NON_INTERACTIVE"] = "1"
    # custom environment variables - this is where cruiz behaviour might vary
    # from Conan on the command line
    env.update(added_environment)
    return (env, removed_environment)



#!/usr/bin/env python3

"""
Dialog that represent progress of operations
"""

import typing

from qtpy import QtCore, QtWidgets

from cruiz.commands.context import ConanContext


class _ContextProgressDialog(QtWidgets.QProgressDialog):
    """
    Common base class for all progress dialogs.
    """

    def __init__(
        self, context: ConanContext, title: str, parent: QtWidgets.QWidget
    ) -> None:
        super().__init__(title, "Cancel", 0, 0, parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        self._title = title
        self._context = context
        self.setValue(0)
        self.canceled.connect(self._cancel)

    def _done(self, result: typing.Any, exception: typing.Any) -> None:
        # pylint: disable=unused-argument
        if exception:
            QtWidgets.QMessageBox.critical(
                self, f"{self._title} failed", str(exception)
            )
            return
        self.reset()

    def _cancel(self) -> None:
        self._context.cancel()


class RemoveLocksDialog(_ContextProgressDialog):
    """
    Progress dialog for removing locks from the Conan local cache.
    """

    def __init__(self, context: ConanContext, parent: QtWidgets.QWidget) -> None:
        super().__init__(context, "Removing locks", parent)
        context.remove_local_cache_locks(self._done)


class RemoveAllPackagesDialog(_ContextProgressDialog):
    """
    Progress dialog for removing all packages from the local cache.
    """

    def __init__(self, context: ConanContext, parent: QtWidgets.QWidget) -> None:
        super().__init__(context, "Removing all packages", parent)
        context.remove_all_packages(self._done)



#!/usr/bin/env python3

"""
QLineEdit subclass that has a custom context menu
"""

from qtpy import QtGui, QtWidgets


class LineEditWithCustomContextMenu(QtWidgets.QLineEdit):
    """
    Subclass allowing a custom context menu to be appended to the standard menu.
    """

    def set_custom_menu(self, menu: QtWidgets.QMenu) -> None:
        """
        Set the custom menu on this widget
        """
        # pylint: disable=attribute-defined-outside-init
        self._custom_menu = menu

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent) -> None:
        menu = self.createStandardContextMenu()
        if self._custom_menu:
            menu.addSeparator()
            menu.addMenu(self._custom_menu)
        menu.exec_(event.globalPos())



#!/usr/bin/env python3

"""
A list, guarded by a mutex, that periodically flushes the joined list
to a QPlainTextEdit
"""

import typing

from qtpy import QtCore, QtWidgets


class GuardedListToFlush(QtCore.QObject):
    """
    Guard a list of strings by a mutex, and flush it to a QPlainTextEdit periodically
    """

    def __init__(self) -> None:
        super().__init__()
        self._widget: typing.Optional[QtWidgets.QPlainTextEdit] = None
        self._list: typing.List[str] = []
        self._guard = QtCore.QMutex()
        self.poll = QtCore.QTimer(self)
        self.poll.timeout.connect(self._flush_list)

    def append(self, message: str) -> None:
        """
        Append a new message to the list.
        """
        # blocking lock
        self._guard.lock()
        self._list.append(message)
        self._guard.unlock()

    def start(self, frequency: int, widget: QtWidgets.QPlainTextEdit) -> None:
        """
        Start flushing the list
        """
        self._widget = widget
        self.poll.start(frequency)

    def stop(self) -> None:
        """
        Stop flushing the list
        """
        self.poll.stop()
        if self._widget:
            # ensure there's no other data left in the buffer
            self._flush_list()

    def _flush_list(self) -> None:
        if not self._list:
            # there may be a thread writing to it, but just wait until the next flush
            return
        if not self._guard.tryLock():
            # give up if writing is happening simultaneously
            return
        combined = "<br>".join(self._list)
        self._list.clear()
        self._guard.unlock()
        assert self._widget
        self._widget.appendHtml(combined)



#!/usr/bin/env python3

"""
Plain old data classes
"""

from __future__ import annotations

from dataclasses import dataclass, field
import os
import typing


# copied from distutils.url.strtobool and modified
def _strtobool(val: str) -> bool:
    """Convert a string representation of truth to true (1) or false (0).

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError("invalid truth value %r" % (val,))


@dataclass(frozen=True)
class LocalCacheDetails:
    """
    Plain old data class representing Conan local cache details.
    """

    home_directory: typing.Optional[str] = None
    short_home_directory: typing.Optional[str] = None
    extra_profile_directories: typing.List[ExtraProfileDirectory] = field(
        default_factory=list
    )
    environment: typing.Dict[str, str] = field(default_factory=dict)
    # TODO: the following are deprecated and will be removed in future
    v2_mode: typing.Optional[bool] = None
    use_revisions: typing.Optional[bool] = None


@dataclass(frozen=True)
class ConanRemote:
    """
    Plain old data class representing a Conan remote.
    """

    name: str
    url: str
    enabled: bool

    @classmethod
    def from_string(cls, string: str) -> ConanRemote:
        """
        Convert a string into a ConanRemote.
        """
        assert string.startswith("ConanRemote(")
        assert string.endswith(")")
        string = string.replace("ConanRemote(", "")
        string = string.replace(")", "")
        args = string.split(",")
        assert len(args) == 3
        name_arg = args[0].strip().split("=")
        url_arg = args[1].strip().split("=")
        enabled_arg = args[2].strip().split("=")
        name = name_arg[1][1:-1]
        url = url_arg[1][1:-1]
        enabled = _strtobool(enabled_arg[1])
        return cls(name, url, enabled)


@dataclass(frozen=True)
class ConanHook:
    """
    Plain old data class representing a Conan hook
    """

    path: str
    enabled: bool

    def has_path(self, path: str) -> bool:
        """
        Does the hook contain the specified path?
        """
        return self.path in (path, os.path.splitext(path)[0])

    @classmethod
    def from_string(cls, string: str) -> ConanHook:
        """
        Convert a string to a ConanHook.
        """
        assert string.startswith("ConanHook(")
        assert string.endswith(")")
        string = string.replace("ConanHook(", "")
        string = string.replace(")", "")
        args = string.split(",")
        assert len(args) == 2
        path_arg = args[0].strip().split("=")
        enabled_arg = args[1].strip().split("=")
        path = path_arg[1][1:-1]
        enabled = _strtobool(enabled_arg[1])
        return cls(path, enabled)


@dataclass(frozen=True)
class ExtraProfileDirectory:
    """
    Plain old data class representing an additional profile directory.
    """

    name: str
    directory: str



#!/usr/bin/env python3

"""
Settings context manager for local cache preferences
"""

import typing

from .basesettings import (
    BaseSettings,
    StringSetting,
    ComparableCommonSettings,
    SettingMeta,
)
from .writermixin import _WriterMixin
from .valueclasses import ScalarValue


# TODO: rename to NewLocalCacheSettings?
class LocalCacheSettings(ComparableCommonSettings):
    """
    Representation of settings for new local caches
    """

    def __init__(self) -> None:
        self._property_meta = {
            "new_configuration_install": SettingMeta(
                "NewConfigInstall", StringSetting, None, ScalarValue
            ),
            "new_configuration_git_branch": SettingMeta(
                "NewConfigGitBranch", StringSetting, None, ScalarValue
            ),
        }

    def presets(self) -> typing.Dict[str, typing.Dict[str, str]]:
        """
        Generate a dictionary of QSettings paths and their values for presets
        """
        group_presets = {}
        new_config = self.new_configuration_install
        if new_config.resolve():
            group_presets["NewConfigInstall"] = new_config.resolve()
        new_config_branch = self.new_configuration_git_branch
        if new_config_branch.resolve():
            group_presets["NewConfigGitBranch"] = new_config_branch.resolve()

        presets = {}
        if group_presets:
            assert self.settings_reader
            presets[self.settings_reader.group] = group_presets
        return presets

    @property
    def new_configuration_install(self) -> StringSetting:
        """
        Get the Conan configuration URL to install for all new local caches
        """
        return self._get_value_via_meta()

    @new_configuration_install.setter
    def new_configuration_install(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def new_configuration_git_branch(self) -> StringSetting:
        """
        Get the Git branch of the Conan configuration to install
        for all new local caches
        """
        return self._get_value_via_meta()

    @new_configuration_git_branch.setter
    def new_configuration_git_branch(self, value: str) -> None:
        self._set_value_via_meta(value)


class LocalCacheSettingsReader:
    """
    Context manager for reading new local cache settings from disk
    """

    def __init__(self) -> None:
        self.group = "LocalCache"
        self.settings = BaseSettings.make_settings()

    def __enter__(self) -> LocalCacheSettings:
        self.settings.beginGroup(self.group)
        self._settings_object = LocalCacheSettings()
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endGroup()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class LocalCacheSettingsWriter(_WriterMixin):
    """
    Utility for writing changed new local cache settings to disk
    """

    def __init__(self) -> None:
        self._reader_for_writer = LocalCacheSettingsReader()

    def presets(self, presets: typing.Dict[str, typing.Dict[str, str]]) -> None:
        """
        Take the dictionary of presets and write them to settings.
        There are no checks that these keys are still valid, so importing
        old presets might cause obsolete keys to be saved with no effect.
        """

        if self._reader_for_writer.group in presets:
            with BaseSettings.Group(self._reader_for_writer.group) as settings:
                for key, value in presets[self._reader_for_writer.group].items():
                    settings.setValue(key, value)



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.packagebinaryparameters import PackageBinaryParameters

from .utils import worker
from .utils.message import Message, Success


def invoke(
    queue: multiprocessing.Queue[Message], params: PackageBinaryParameters
) -> None:
    """
    Similar to 'conan download' but the download folder is custom.

    PackageBinaryParameters has dynamic attributes.
    """
    # pylint: disable=import-outside-toplevel, protected-access, no-member
    from conans.model.ref import PackageReference

    with worker.ConanWorker(queue, params) as api:
        try:
            # Conan 1.18+
            remote_manager = api.app.remote_manager
        except AttributeError:
            remote_manager = api._remote_manager

        remote_name = params.remote_name  # type: ignore
        pkgref = params.reference  # type: ignore
        where = params.where  # type: ignore

        try:
            # Conan 1.19+
            remotes = api.app.load_remotes(remote_name=remote_name)
        except AttributeError:
            try:
                # Conan 1.18.x
                remotes = api.app.cache.registry.load_remotes()
            except AttributeError:
                remotes = api._cache.registry.load_remotes()
        remote = remotes.get_remote(remote_name)
        pref_str = pkgref
        pref = PackageReference.loads(pref_str)
        # from RemoteManager.get_package
        snapshot = remote_manager._call_remote(remote, "get_package_snapshot", pref)
        remote_manager._call_remote(remote, "get_package", pref, where)

        queue.put(Success(snapshot))



#!/usr/bin/env python3

"""
Wrapper around a child process, that can process multiple requests for
meta data from a Conan instance
"""

import logging
import multiprocessing
import os
import sys
import typing
import urllib.parse

from qtpy import QtCore

from cruiz.interop.commandparameters import CommandParameters
import cruiz.workers.utils.message
import cruiz.workers.meta
from cruiz.dumpobjecttypes import dump_object_types

from .conanenv import get_conan_env
from .logdetails import LogDetails


logger = logging.getLogger(__name__)


class MetaRequestConanInvocation(QtCore.QObject):
    """
    Wrapper around request-reply interaction for meta data from a Conan local cache
    Request-replies are synchronous, since it's just metadata
    """

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(  # noqa: F811
        self,
        parent: QtCore.QObject,
        cache_name: str,
        log_details: LogDetails,
    ) -> None:
        logger.debug("+=%d", id(self))
        super().__init__(parent)
        self._log_details = log_details
        self._mp_context = multiprocessing.get_context("spawn")
        self._request_queue = self._mp_context.JoinableQueue()
        self._reply_queue = self._mp_context.Queue()
        self.active = False
        params = CommandParameters("meta", cruiz.workers.meta.invoke)
        added_environment, removed_environment = get_conan_env(cache_name)
        params.added_environment.update(added_environment)
        params.removed_environment.extend(removed_environment)
        self._invoke_conan_process(params)

    def close(self) -> None:
        """
        Close all resources associated with the invocation
        """
        logger.debug("(%d) closing request queue...", id(self))
        self._request_queue.put("end")
        self._request_queue.join()
        self._request_queue.close()
        self._request_queue.join_thread()
        logger.debug("(%d) closing reply queue...", id(self))
        self._reply_queue.close()
        self._reply_queue.join_thread()
        logger.debug("(%d) joining process...", id(self))
        self._process.join()
        self._process.close()

    def _invoke_conan_process(self, params: CommandParameters) -> None:
        process = self._mp_context.Process(
            target=params.worker,
            args=(self._request_queue, self._reply_queue, params),
            daemon=False,
        )
        process.start()
        logger.debug(
            "cruiz (pid=%i) started child process (pid=%i) for %s",
            os.getpid(),
            process.pid,
            params.worker.__module__,
        )
        self._process = process

    @staticmethod
    def __check_for_conan_leakage(entry: typing.Any = None) -> None:
        if "conans" not in sys.modules:
            return
        if entry:
            logger.critical("Conan has leaked into message queue object dump:")
            dump_object_types(entry, loglevel="CRITICAL")
        logger.critical("Conan has leaked into cruiz")
        sys.exit(1)

    def request_data(
        self,
        request: typing.Any,
        params: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> typing.Tuple[typing.Any, typing.Optional[Exception]]:
        """
        Request a named metadata, and synchronously wait for a reply.
        Reply is a tuple, with result and exception.
        """
        meta_request = (
            f"{request}?{urllib.parse.urlencode(params, doseq=True)}"
            if params
            else request
        )
        logger.debug("* Requesting %s", meta_request)
        self._request_queue.put(meta_request)
        self.active = True
        logger.debug("* Requested %s waiting for reply", meta_request)
        # the requested task is done when there is a success or a failure
        self._request_queue.join()
        response: typing.Union[
            cruiz.workers.utils.message.Success,
            cruiz.workers.utils.message.Failure,
            None,
        ] = None
        while True:
            MetaRequestConanInvocation.__check_for_conan_leakage(None)
            reply = self._reply_queue.get()
            MetaRequestConanInvocation.__check_for_conan_leakage(reply)
            if isinstance(reply, cruiz.workers.utils.message.Stdout):
                logger.debug("* Got stdout message: '%s", reply.message())
                self._log_details.stdout(reply.message())
            elif isinstance(reply, cruiz.workers.utils.message.Stderr):
                logger.debug("* Got stderr message: '%s", reply.message())
                self._log_details.stderr(reply.message())
            elif isinstance(reply, cruiz.workers.utils.message.ConanLogMessage):
                logger.debug("* Got Conan log message: '%s", reply.message())
                self._log_details.conan_log(reply.message())
            elif isinstance(reply, cruiz.workers.utils.message.Success):
                logger.debug("* Requested %s got reply '%s'", meta_request, reply)
                response = reply
                break
            elif isinstance(reply, cruiz.workers.utils.message.Failure):
                logger.debug("* Got failure message: '%s'", str(reply.exception()))
                response = reply
                break
        assert self._reply_queue.empty()
        self.active = False
        if response is not None:
            if isinstance(reply, cruiz.workers.utils.message.Success):
                return (response.payload(), None)  # type: ignore
            if isinstance(reply, cruiz.workers.utils.message.Failure):
                return (None, response.exception())  # type: ignore
        raise RuntimeError("No success message")



#!/usr/bin/env python3

"""
Worker class that is run on a background thread, processing replies
from child processes and converting to signals.
"""

from __future__ import annotations

import logging
import multiprocessing
import sys
import typing

from qtpy import QtCore

from cruiz.dumpobjecttypes import dump_object_types

import cruiz.workers.endmessagethread
import cruiz.workers.utils.message

logger = logging.getLogger(__name__)


class MessageReplyProcessor(QtCore.QObject):
    """
    Process replies across a multiprocessing Queue in a background thread.
    Emit signals for recognised messages, since you cannot add to UI in a
    background thread.
    """

    completed = QtCore.Signal(object, Exception)
    stdout_message = QtCore.Signal(str)
    stderr_message = QtCore.Signal(str)
    conan_log_message = QtCore.Signal(str)
    critical_failure = QtCore.Signal(str)

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(
        self,
        queue: multiprocessing.Queue[cruiz.workers.utils.message.Message],
    ):
        logger.debug("+=%d", id(self))
        super().__init__()
        self._mp_context = multiprocessing.get_context("spawn")
        self._queue = queue

    def stop(self) -> None:
        """
        Stop the background thread.
        """
        # start a process to shut down the thread
        shutdown_process = self._mp_context.Process(
            target=cruiz.workers.endmessagethread.invoke, args=(self._queue,)
        )
        shutdown_process.start()
        shutdown_process.join()
        shutdown_process.close()

    def __check_for_conan_leakage(self, entry: typing.Any = None) -> bool:
        if "conans" not in sys.modules:
            return True
        if entry:
            logger.critical("Conan has leaked into message queue object dump:")
            dump_object_types(entry, loglevel="CRITICAL")
            raise AssertionError("Conan has leaked into cruiz")
        self.critical_failure.emit("Conan has leaked into cruiz")
        return False

    # pylint: disable=too-many-branches
    def process(self) -> None:
        """
        Process messages received from a child process.
        """
        try:
            # pylint: disable=import-error,import-outside-toplevel
            # make any debugger aware of this thread
            import pydevd

            pydevd.settrace(suspend=False)
        except ModuleNotFoundError:
            pass
        while True:
            logger.debug("(%d) wait for queue entry...", id(self))
            try:
                if not self.__check_for_conan_leakage(None):
                    break
                entry = self._queue.get()
                if not self.__check_for_conan_leakage(entry):
                    break
                if isinstance(entry, cruiz.workers.utils.message.End):
                    break
                if isinstance(entry, cruiz.workers.utils.message.Stdout):
                    self.stdout_message.emit(entry.message())
                elif isinstance(entry, cruiz.workers.utils.message.Stderr):
                    self.stderr_message.emit(entry.message())
                elif isinstance(entry, cruiz.workers.utils.message.ConanLogMessage):
                    self.conan_log_message.emit(entry.message())
                elif isinstance(entry, cruiz.workers.utils.message.Success):
                    self.completed.emit(entry.payload(), None)
                elif isinstance(entry, cruiz.workers.utils.message.Failure):
                    # TODO: temporary, at least always record the exception
                    # in the error log
                    self.stderr_message.emit(str(entry.exception()))
                    self.completed.emit(None, entry.exception())
                else:
                    logger.error("Unknown message type: '%s'", entry)
            except EOFError as exception:
                logger.debug(
                    "(%d) queue failed because %s",
                    id(self),
                    str(exception),
                )
        logger.debug("(%d) closing queue...", id(self))
        self._queue.close()
        self._queue.join_thread()
        logger.debug("(%d) closed queue", id(self))
        QtCore.QThread.currentThread().quit()



#!/usr/bin/env python3

"""
Helpful widget classes and utilities
"""

from cruiz.widgets.shortcutlineedit import ShortcutLineEdit
from cruiz.widgets.aboutcruizdialog import AboutDialog
from cruiz.widgets.debugging import log_created_widget



#!/usr/bin/env python3

"""
Global variables
(could do without these)
"""

from __future__ import annotations
import typing

CRUIZ_MAINWINDOW: typing.Optional[cruiz.MainWindow] = None  # type: ignore # noqa: F821


def get_main_window() -> cruiz.MainWindow:  # type: ignore # noqa: F821
    """
    Get the main window for the application, asserting it is valid
    """
    assert CRUIZ_MAINWINDOW
    return CRUIZ_MAINWINDOW



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan remove --locks'
    """
    with worker.ConanWorker(queue, params) as api:
        result = api.remove_locks()

        queue.put(Success(result))
